import numpy as np
import numexpr as ne


__author__  = "Davide Micieli"
__all__     = ['get_circular_mask', 'circular_crop']

def get_circular_mask(nrow, ncol, radius=None, center=None):
	"""
	This function returns a boolean 2D array representing a circular mask.
	The values outside the circle are `False`, while the inner values are `True`.

	Parameters
	----------
	nrow: int
		The number of rows of the mask array.

	ncol: int
		The number of columns of the mask array.

	radius: float, optional
		The radius of the circle in pixel.
		The default value is 0.5 * min(number of rows,number of columns)

	center: tuple of float, optional (yc, xc)
		A tuple representing the coordinates of the center of the circle, i.e.: (yc, xc).
		The default value is the center of the input image.

	Returns
	-------
	mask : 2d array
		A boolean array that represents the circular mask. The values outside the circle
		are `False`, while the inner values are `True`.
	"""


	if(radius is None):
		radius = min(ncol, nrow)/2

	if(center is None):
		yc = ncol/2.0
		xc = nrow/2.0
	else:
		yc, xc = center

	ny = np.arange(ncol)
	nx = np.arange(nrow)

	x, y = np.meshgrid(nx, ny)

	mask = ( (y-yc + 0.5)**2 + (x-xc + 0.5)**2 ) < (radius)**2

	return mask



def circular_crop(img, axis=0, radius=None, center=None, cval=0):
	"""
	This function performs a circular crop of an image. The values outside the
	circle are replaced with the value cval (default cval=0).

	Parameters
	----------

	img: 2d array
		The array representing the image to crop.

	radius: float, optional
		The radius of the circle. The default value is 0.5 * min(number of rows,number of columns)

	center: tuple of float, optional (yc, xc)
		A tuple representing the coordinates of the center of the circle, i.e.: (yc, xc).
		The default value is the center of the input image.

	cval: float, optional
		The value used to fill the region outside the circle.

	Returns
	-------

	out: 2d array
		The cropped image.
	"""

	if (img.ndim == 3):
		img = img.swapaxes(0,axis)
		nslice, nrow, ncol = img.shape
	elif (img.ndim == 2):
		nrow, ncol = img.shape
	else:
		raise ValueError("The input array must have 2 or 3 dimensions.")

	# set type of cval
	cval =	np.array(cval, dtype=img.dtype)

	# get the circular mask	-> True inside the circle
	mask = get_circular_mask(nrow, ncol, radius, center)

	# apply the mask
	ne.evaluate('where(mask, img, cval)' , out=img)

	if (img.ndim == 3):
		img = img.swapaxes(0,axis)

	return img



import numpy as np
import logging

__author__  = "Davide Micieli"
__all__     = ['rebin']


logs = logging.getLogger(__name__)


def rebin3D(arr, binsize, dtype=np.float32):
	"""
	This function rebins a 3D stack of images by averaging.

	Parameters
	----------
	arr :  ndarray
		The 3D stack of images to rebin.

	binsize: tuple
		Tuple defining the bin size for each axis of the 3D array.

	dtype: type, optional
		The type of the rebbined array.

	Returns
	-------
	m : ndarray
		The rebinned stack.
	"""
	arr = arr.astype(np.float32)

	shape = ( arr.shape[0] // binsize[0], binsize[0],
			  arr.shape[1] // binsize[1], binsize[1],
			  arr.shape[2] // binsize[2], binsize[2])

	resto_0 = arr.shape[0]  % binsize[0]
	resto_1 = arr.shape[1]  % binsize[1]
	resto_2 = arr.shape[2]  % binsize[2]

	if  (resto_0 != 0):
		arr = arr[:-resto_0, :, :]

	if(resto_1 != 0):
		arr = arr[:, :-resto_1, :]

	if(resto_2 != 0):
		arr = arr[:, : , :-resto_2]

	return arr.reshape(shape).mean(-1).mean(1).mean(2).astype(dtype)


def rebin2D(arr, binsize, dtype=np.float32):
	"""
	This function rebins a 2D image by averaging.

	Parameters
	----------
	arr :  ndarray
		The 2D image to rebin.

	binsize: tuple
		Tuple defining the bin size for each axis of the image.

	dtype: type, optional
		The type of the rebbined array.

	Returns
	-------
	m : ndarray
		The rebinned image.
	"""

	arr = arr.astype(np.float32)

	shape = ( arr.shape[0] // binsize[0], binsize[0],
		      arr.shape[1] // binsize[1], binsize[1])

	resto_0 = arr.shape[0]  % binsize[0]
	resto_1 = arr.shape[1]  % binsize[1]

	if  (resto_0 != 0):
		arr = arr[:-resto_0, :]

	if(resto_1 != 0):
		arr = arr[:, :-resto_1]

	return arr.reshape(shape).mean(-1).mean(1).astype(dtype)


def rebin(arr, binsize, dtype=np.float32):
	"""
	This function rebins a single 2D image or 3D stack.
	It reduces the size of an image or stack of images by binning groups of pixels of user-specified sizes. The resulting pixels are computed as average.

	Parameters
	----------
	arr :  ndarray, 2d or 3d
		The 2D image or the 3D stack of images to rebin.

	binsize: tuple
		Tuple defining the bin size for each axis of the array.
		E.g.: for 2d array (bin_0, bin_1)
			  for 3d array (bin_0, bin_1, bin_2)

	dtype: type, optional
		The type of the returned rebbined array.

	Returns
	-------
	m : ndarray
		The rebinned image or stack of images.
	"""
	if(arr.ndim==3 and len(binsize)==3):
		return rebin3D(arr, binsize, dtype)

	elif (arr.ndim==2 and len(binsize)==2):
		return rebin2D(arr, binsize, dtype)

	else:
		raise ValueError("Array shape or binning shape not valid.")



import numpy as np
import tifffile
import astropy
from astropy.io import fits
from read_roi import read_roi_file
import logging
import os
import types
import inspect
from neutompy.misc.uitools import get_filename_gui, get_image_gui, get_folder_gui
from tqdm import tqdm

__author__  = "Davide Micieli"
__all__     = ['read_tiff',
			 'read_fits',
			 'read_fits_stack',
			 'read_tiff_stack',
			 'get_rect_coordinates_from_roi',
			 'read_image',
			 'read_stack_from_list',
			 'read_dataset',
			 'read_image_stack',
			 'write_fits',
			 'write_tiff',
			 'write_tiff_stack',
			 'write_fits_stack',
			 'get_rect_coordinates_from_roi', 'get_filename_pattern']

#logging.basicConfig(level=logging.WARNING)
logs = logging.getLogger(__name__)


known_ext = ['.tif', '.tiff', '.fits']

# to ensure compatibility with astropy 1.X, clobber option is deprecated and replaced with overwrite
if 'clobber' in inspect.getfullargspec(astropy.io.fits.writeto).args:
	arg_overwrite = 'clobber'
else:
	arg_overwrite = 'overwrite'


def get_file_extension(fname):
	"""
	This function returns the extension of a given file name or file path
	"""
	return os.path.splitext(fname)[1]


def get_rect_coordinates_from_roi (fname):
	"""
		This function returns the coordinates from a rectangular region of interest
		defined in a .roi file generated by ImageJ / Fiji.

		N.B.: indexing starts from 0.
		Incremet rowmax and col_max  + 1 to crop an image, for example:
		crop = img[rowmin:(rowmax+1),  colmin:(colmax+1)]

		Parameters
		----------
		fname : str
			String defining the path or the name of the Image ROI file.

		Returns
		-------
		rowmin : int
			The minimum row coordinate.

		rowmax : int
			The maximum row coordinate.

		colmin : int
			The minimum column coordinate.

		colmax : int
			The maximum column coordinate.

	"""

	# file name check
	if(type(fname) is not str):
		logs.error('File name must be a string.')
		raise TypeError('File name must be a string.')
	else:
		if not(fname.endswith('.roi')):
			logs.warning('File extension shold be .roi')

	# read roi file
	roi = read_roi_file(fname)


	# convert keys to list
	l = list(roi.keys())

	# extract coordinates and length
	height = roi[l[0]]['height']
	width  = roi[l[0]]['width']
	top    = roi[l[0]]['top']
	left	 = roi[l[0]]['left']

	rowmin = int(top)
	rowmax = int(top  + height - 1)
	colmin = int(left)
	colmax = int(left + width - 1)

	# log infos
	logs.info('ROI file succesfully read: %s', fname)

	return rowmin, rowmax, colmin, colmax


def read_tiff(fname, croi=None, froi=None):
	"""
	This function reads a 2D TIFF image.

	Parameters
	----------
	fname : str
		String defining the file name or the file path

	croi : tuple, optional
		Tuple defining the indexes range for each axis. It must be follow this notation:
		( (row_start, row_end, row_step),  (col_start, col_end, col_step) )

	froi : str, optional
		String defining the ImageJ ROI file name or file path.

	Returns
	-------
	out : ndarray
		The 2D image.
		If the file specified doesn't exist, the function returns False.

	"""
	# file name check
	if(type(fname) is not str):
		logs.error('File name must be a string.')
		raise TypeError('File name must be a string.')

	if ( not(fname.endswith('.tif')) and  not(fname.endswith('.tiff')) ):
		logs.warning('File extension shold be .tif or .tiff')

	# ROI check
	if(croi and froi):
		raise ValueError('Please set the coordinates of the region of interest OR the ImageJ file .roi')

	if(croi):
		if not(len(croi)==2 and len(croi[0])==3  and len(croi[1])==3  ):
			raise ValueError('Tuple or list size not valid. Please set the coordinates of the ROI in the this manner:   croi = ( (row_start, row_end, row_step),  (col_start, col_end, col_step) ) ')


	# get the absolute path of the file
	fname = os.path.abspath(fname)

	try:
		out = tifffile.imread(fname) #, out='memmap')
		logs.info('Array imported from file: %s', fname)
		logs.debug('Array shape and type: %s, %s', out.shape, out.dtype)
	except IOError:
		logs.error('No such file: %s', fname)
		return False

	if (len(out.shape) != 2):
		raise ValueError('Only 2D TIFF images are supported in read_tiff')

	if(croi):
		rstart, rend, rstep = croi[0]
		cstart, cend, cstep = croi[1]
		if (rend is not None):
			rend = rend + 1
		if (cend is not None):
			cend = cend + 1
		out  = out[rstart:rend:rstep, cstart:cend:cstep]

		logs.info('Read roi coordinate from list / tuple')
		logs.debug('Cropped array shape and type: %s, %s', out.shape, out.dtype)
		logs.debug('rstart =  %s, rend = %s, rstep = %s', rstart, rend, rstep)
		logs.debug('cstart =  %s, cend = %s, cstep = %s', cstart, cend, cstep)

	if(froi):
		rstart, rend, cstart, cend = get_rect_coordinates_from_roi(froi)
		out = out[rstart:rend + 1, cstart:cend + 1]

		logs.info('Read ROI coordinate from file : %s' , froi)
		logs.debug('Cropped array shape and type: %s, %s', out.shape, out.dtype)
		logs.debug('rstart =  %s, rend = %s, rstep = %s', rstart, rend)
		logs.debug('cstart =  %s, cend = %s, cstep = %s', cstart, cend)


	return out

def read_fits(fname, croi=None, froi=None):
	"""
	This function reads a 2D FITS image.

	Parameters
	----------
	fname : str
		String defining the file name or the file path

	croi : tuple, optional
		Tuple defining the indexes range for each axis. It must be follow this notation:
		( (row_start, row_end, row_step),  (col_start, col_end, col_step) )

	froi : str, optional
		String defining the ImageJ ROI filename or file path.

	Returns
	-------
	out : ndarray
		The 2D image.
		If the file specified doesn't exist, the function returns False.

	"""
	# file name check
	if(type(fname) is not str):
		logs.error('File name must be a string.')
		raise TypeError('File name must be a string.')

	if ( not(fname.endswith('.fits')) ):
		logs.warning('File extension shold be .fits')

	# ROI check
	if(croi and froi):
		raise ValueError('Please set the coordinates of the region of interest OR the ImageJ file .roi')

	if(croi):
		if not(len(croi)==2 and len(croi[0])==3  and len(croi[1])==3  ):
			raise ValueError('Tuple or list size not valid. Please set the coordinates of the ROI in the this manner:   croi = ( (row_start, row_end, row_step),  (col_start, col_end, col_step) ) ')




	# get the absolute path of the file
	fname = os.path.abspath(fname)

	try:
		f = fits.open(fname)
		out = f[0].data[::-1, :].copy() # flip up-down
		f.close()
		logs.info('Array imported from file: %s', fname)
		logs.debug('Array shape and type: %s, %s', out.shape, out.dtype)
	except IOError:
		logs.error('No such file: %s', fname)
		return False

	if (len(out.shape) != 2):
		raise ValueError('Only 2D FITS images are supported in read_tiff')

	if(croi):
		rstart, rend, rstep = croi[0]
		cstart, cend, cstep = croi[1]
		if (rend is not None):
			rend = rend + 1
		if (cend is not None):
			cend = cend + 1
		out  = out[rstart:rend:rstep, cstart:cend:cstep]

		logs.info('Read roi coordinate from list / tuple')
		logs.debug('Cropped array shape and type: %s, %s', out.shape, out.dtype)
		logs.debug('rstart =  %s, rend = %s, rstep = %s', rstart, rend, rstep)
		logs.debug('cstart =  %s, cend = %s, cstep = %s', cstart, cend, cstep)

	if(froi):
		rstart, rend, cstart, cend = get_rect_coordinates_from_roi(froi)
		out = out[rstart:rend + 1, cstart:cend + 1]

		logs.info('Read ROI coordinate from file : %s' , froi)
		logs.debug('Cropped array shape and type: %s, %s', out.shape, out.dtype)
		logs.debug('rstart =  %s, rend = %s, rstep = %s', rstart, rend)
		logs.debug('cstart =  %s, cend = %s, cstep = %s', cstart, cend)

	return out



def read_image(fname, croi=None, froi=None):
	"""
	This function reads a 2D TIF or FITS image.

	Parameters
	----------
	fname : str
		String defining the file name or the file path

	croi : tuple, optional
		Tuple defining the indexes range for each axis. It must be follow this notation:
		( (row_start, row_end, row_step),  (col_start, col_end, col_step) )

	froi : str, optional
		String defining the ImageJ ROI filename or file path.

	Returns
	-------
	out : ndarray
		The 2D image.
		If the file specified doesn't exist, the function returns False.

	"""

	# check if file exists
	if not(os.path.isfile(fname)):
		raise ValueError('No such file %s' %fname)

	# read .fits or .tif from the file extension
	if(fname.endswith('.fits') or fname.endswith('.FITS')):
		funz = read_fits
	elif(fname.endswith('.tif') or fname.endswith('.tiff') or fname.endswith('.TIF') or fname.endswith('.TIFF')):
		funz = read_tiff
	else:
		raise ValueError('File extension not valid. File : %s' % fname)

	array = funz(fname, croi, froi)

	return array



def get_filename_pattern(fname):
	# get file name extension
	ext = get_file_extension(fname)

	# get prefix_XXXX
	aus = fname[:-len(ext)]

	# find how many X, remove XXXX and find the prefix name
	p = 0
	while(aus[-1-p].isdigit()):
		p = p + 1
	prefix_full = aus[:-p]

	folder = os.path.dirname(fname)
	prefix = os.path.basename(prefix_full)

	numbering = '#'*p

	return prefix_full, numbering, ext


def read_stack_from_list(flist, slices=[], croi=None, froi=None):
	"""
	Read stack of images from file list.

	Parameters
	----------
	flist : list of str
		List of the file names or file paths to read in the defined order. File with unknown image extension are skipped automatically.

	slices : list of int, optional
		List of the indexes of flist to read. If it is [] then all image files are read.

	croi : tuple, optional
		Tuple defining the indexes range for each axis. It must be follow this notation:
		( (row_start, row_end, row_step),  (col_start, col_end, col_step) )

	froi : str, optional
		String defining the ImageJ ROI filename or file path.

	Returns
	-------
	array : ndarray
		The 3D stack of images

	"""

	# if slices == [], initialize all indexes
	if not slices:
		slices = [ i for i in range(0, len(flist)) ]

	# list of index to read
	toread = [True]*len(slices)

	# skip files with unknown extension
	for e, s in enumerate(slices):
		if not (get_file_extension(flist[s]) in known_ext):
			toread[e] = False
			logs.warning('Unknown file extension. Skipped file: %s', flist[s])


	fpat = get_filename_pattern(flist[0])  # e.g. /folder1/folder2/img_####.tiff

	print('> Reading stack of images:', ''.join(fpat))
	array = np.array([read_image(flist[s], croi, froi) for e, s in enumerate(tqdm(slices, unit=' images')) if toread[e]])
	print('')

	return array


def read_tiff_stack(fname, slices=[], croi=None, froi=None):
	"""
	Read stack of tiff images.

	Parameters
	----------
	fname :  str
		One of the file names of the tiff stack. File with unknown image extension are skipped automatically. If fname is '' or a folder path, then a dialog box is opened to select one of the file of the stack to read. The initial directory is C:\ (in windows) or / (in UNIX) if fname=='', otherwise is the folder path assigned to fname.

	slices : list of int, optional
		List of the element indexes to read. If it is [] then all tif image files are read.

	croi : tuple, optional
		Tuple defining the indexes range for each axis. It must be follow this notation:
		( (row_start, row_end, row_step),  (col_start, col_end, col_step) )

	froi : str, optional
		String defining the ImageJ ROI filename or file path.

	Returns
	-------
	array : ndarray
		The 3D stack of images

	"""

	if(os.path.isdir(fname) or fname==''):
		fname = get_filename_gui(initialdir=fname, message='Select a file of the image stack...',
					ext=('TIFF Images', '*.tif *.tiff') )

	# check if file exists
	if not(os.path.isfile(fname)):
		raise ValueError('No such file %s' %fname)

	# get file name extension
	ext = get_file_extension(fname)

	# get prefix_XXXX
	aus = fname[:-len(ext)]

	# find how many X, remove XXXX and find the prefix name
	p = 0
	while(aus[-1-p].isdigit()):
		p = p + 1
	prefix_full = aus[:-p]

	folder = os.path.dirname(fname)
	prefix = os.path.basename(prefix_full)

	flist = [os.path.normpath(os.path.join(folder, f)) for f in sorted(os.listdir(folder)) if (f.startswith(prefix) and f[-p-len(ext):-len(ext)].isdigit() and (get_file_extension(f) in ['.tif', '.tiff']) ) ]

	out = read_stack_from_list(flist, slices, croi, froi)

	return out


def read_fits_stack(fname, slices=[], croi=None, froi=None):
	"""
	Read stack of fits images.

	Parameters
	----------
	fname :  str
		One of the file names of the fits stack. File with unknown image extension are skipped automatically. If fname is '' or a folder path, then a dialog box is opened to select one of the file of the stack to read. The initial directory is C:\ (in windows) or / (in UNIX) if fname=='', otherwise is the folder path assigned to fname.

	slices : list of int, optional
		List of the element indexes to read. If it is [] then all fits image files are read.

	croi : tuple, optional
		Tuple defining the indexes range for each axis. It must be follow this notation:
		( (row_start, row_end, row_step),  (col_start, col_end, col_step) )

	froi : str, optional
		String defining the ImageJ ROI filename or file path.

	Returns
	-------
	array : ndarray
		The 3D stack of images

	"""

	if(os.path.isdir(fname) or fname==''):
		fname = get_filename_gui(initialdir=fname, message='Select a file of the image stack...', ext=('FITS Images', '*.fits') )

	# check if file exists
	if not(os.path.isfile(fname)):
		raise ValueError('No such file %s' %fname)

	# get file name extension
	ext = get_file_extension(fname)

	# get prefix_XXXX
	aus = fname[:-len(ext)]

	# find how many X, remove XXXX and find the prefix name
	p = 0
	while(aus[-1-p].isdigit()):
		p = p + 1
	prefix_full = aus[:-p]

	folder = os.path.dirname(fname)
	prefix = os.path.basename(prefix_full)

	flist = [os.path.normpath(os.path.join(folder, f)) for f in sorted(os.listdir(folder)) if (f.startswith(prefix) and f[-p-len(ext):-len(ext)].isdigit() and (get_file_extension(f) in ['.fits']) ) ]

	out = read_stack_from_list(flist, slices, croi, froi)

	return out

def read_image_stack(fname, slices=[], croi=None, froi=None):
	"""
	Read stack of TIFF or FITS images. This function recognize the file type from
	the file extension.

	Parameters
	----------
	fname :  str
		One of the file names of the tiff stack. File with unknown image extension are
		skipped automatically.

	slices : list of int, optional
		List of the element indexes to read. If it is [] then all tif image files are read.

	croi : tuple, optional
		Tuple defining the indexes range for each axis. It must be follow this notation:
		( (row_start, row_end, row_step),  (col_start, col_end, col_step) )

	froi : str, optional
		String defining the ImageJ ROI filename or file path.

	Returns
	-------
	array : ndarray
		The 3D stack of images

	"""
	if(fname.endswith('.fits') or fname.endswith('.FITS')):
		funz = read_fits_stack
	elif(fname.endswith('.tif') or fname.endswith('.tiff') or fname.endswith('.TIFF') or fname.endswith('.TIF')):
		funz = read_tiff_stack
	else:
		raise ValueError('File extension not valid. File : %s' % fname)

	arr = funz(fname, slices, croi, froi)

	return arr


def read_dataset(proj_180=True, croi=None, froi=None):
	"""
	This function reads a dataset which contains dark-field, flat-field, projection
	images and optionally the projection at 180 degree. The user selects the main folder
	and the files from a dialog box. A two-dimensional region of interest of each
	image can be read specifying the coordinates or an ImageJ .roi file.

	Parameters
	----------
	proj_180 : bool, optional
		If True the user must select the projection at 180 degree separately from
		the stack of projections.

	croi : tuple, optional
		Tuple defining the ROI indexes for each axis. It must be follow this notation:
		( (row_start, row_end, row_step),  (col_start, col_end, col_step) )

	froi : str, optional
		String defining the ImageJ ROI file name or file path.

	Returns
	-------
	proj : 3d array
		The array containing the stack of projections.

	dark : 3d array
		The array containing the stack of dark-field images.

	flat : 3d array
		The array containing the stack of flat-field images.

	proj180 : 2d array
		Only if proj_180 is True, the 2D array representing the projection at 180 degree
		is returned separately.
	"""
	# select the folder that contains dark, flat, and projection images using the GUI
	data_folder = get_folder_gui('', message = 'Select the main folder of the dataset...')

	proj_dir = get_image_gui(data_folder, message = 'Select a raw projection...')
	dark_dir = get_image_gui(data_folder, message = 'Select a dark image...')
	flat_dir = get_image_gui(data_folder, message = 'Select a flat image...')
	if proj_180:
		proj_pi_fname =  get_image_gui(data_folder, message = 'Select raw projection at 180°...')

	# load the dataset
	proj = read_image_stack(proj_dir, croi=croi, froi=froi)
	dark = read_image_stack(dark_dir, croi=croi, froi=froi)
	flat = read_image_stack(flat_dir, croi=croi, froi=froi)
	if proj_180:
		print('> Reading projection at 180 degree:',proj_pi_fname)
		proj180 = read_image(proj_pi_fname, croi=croi, froi=froi)

	if proj_180:
		return proj, dark, flat, proj180
	else:
		return proj, dark, flat


def write_tiff(fname, img, overwrite=False):
	"""
	This function write to the disk an array as tiff image.

	Parameters
	----------
	fname : str
		String defining the file name or file path of the image to save.
		If the extension is not specified, it is automatically appended to fname.

	img : ndarray
		The array to save as image.

	overwrite: bool, optional
		If ``True``, overwrites the output file if it exists. Raises an
		``IOError`` if ``False`` and the output file exists. Default is ``False``.
	"""

	# add extension .tif if not specified in fname
	if not ( fname.endswith('.tif') or fname.endswith('.tiff') ):
		fname = fname + '.tiff'

	if not overwrite:
		if(os.path.isfile(fname)):
			raise IOError("File already exists : %s" %fname)

	if overwrite:
		if(os.path.isfile(fname)):
			logs.info("File to overwrite: %s", fname)

	tifffile.imsave(fname, img)
	logs.info('File saved to disk: %s', fname)
	logs.debug('Image shape %s and type %s', img.shape,  img.dtype)


def write_fits(fname, img, overwrite=False):
	"""
	This function write to the disk an array as fits image.

	Parameters
	----------
	fname : str
		String defining the file name or file path of the image to save.
		If the extension is not specified, it is automatically appended to fname.

	img : ndarray
		The array to save as image.

	overwrite: bool, optional
		If ``True``, overwrites the output file if it exists. Raises an ``IOError``
		if ``False`` and the output file exists. Default is ``False``.
	"""

	# add extension .tif if not specified in fname
	if not ( fname.endswith('.fits') ):
		fname = fname + '.fits'

	if overwrite:
		if(os.path.isfile(fname)):
			logs.info("File to overwrite: %s", fname)

	hdu = fits.PrimaryHDU()
	hdu.data = img[::-1, :]
	hdu.writeto(fname, **{arg_overwrite : overwrite})

	logs.info('File saved to disk: %s', fname)
	logs.debug('Image shape %s and type %s', img.shape,  img.dtype)


def write_tiff_stack(fname, data, axis=0, start=0, croi=None, digit=4, dtype=None, overwrite=False):
	"""
	This function writes a 3D array to a stack of 2D TIFF images.

	Parameters
	----------

	fname: str
		String defining the prefix of the file name and the containing folder.

	data : ndarray
		The 3D stack to write.

	axis : int, optional
		The axis along which the stacking is performed.

	start : int, optional
		Index used for saving the first image.

	croi : tuple, optional
		Tuple defining the indexes range for each axis. It must be follow this notation:
		( (row_start, row_end, row_step),  (col_start, col_end, col_step) )

	digit : int, optional
		Number of digits used for the numbering of the images.

	dtype : dtype, optional
		Data type of the images to save.

	overwrite: bool, optional
		If ``True``, overwrites the output file if it exists. Raises an ``IOError``
		if ``False`` and the output file exists. Default is ``False``.

	"""
	## check dimension data ==3
	if (data.ndim != 3):
		raise ValueError('Array must have 3 dimensions.')

	folder = os.path.dirname(fname)

	if folder == '':
		raise ValueError('File path not valid. Please specify the file path giving at\
				least one folder, e.g.: folder/file.tiff ')
	else:
		if not (os.path.isdir(folder)):
			raise ValueError('Folder not exist. %s' %folder)

	if not dtype:
		dtype = data.dtype


	if(axis!=0):
		data  = np.swapaxes(data, 0, axis)
		if(croi):
			newcroi_list       = [[None]*3]*3
			newcroi_list[axis] = croi[0]
			newcroi_list[0]    = croi[axis]
			newcroi_list[3-axis] = croi[3-axis]
			newcroi              = tuple(newcroi_list)

	if not croi:
		slices, rows, cols = data.shape
		newcroi = ((0, slices, 1), (0, rows, 1), (0, cols, 1))
	else:
		if(axis ==0):
			newcroi = croi

	rmin, rmax, rstep = newcroi[1]
	cmin, cmax, cstep = newcroi[2]
	smin, smax, sstep = newcroi[0]

	auslist = list(range(0, data.shape[0], 1))

	subset = auslist[smin:smax:sstep]

	logs.debug('Shape of the re-stacked array %s',  newcroi)

	print('> Saving stack of images:', fname + '_' + '#'*digit + '.tiff')
	for iz, s in enumerate(tqdm(subset, unit=' images'), start):
		outfile = fname + '_' + str(iz).zfill(digit) + '.tiff'
		write_tiff(outfile, data[s, rmin:rmax:rstep, cmin:cmax:cstep].astype(dtype), overwrite)


def write_fits_stack(fname, data, axis=0, start=0, croi=None, digit=4, dtype=None, overwrite=False):
	"""
	This function writes a 3D array to a stack of 2D FITS images.

	Parameters
	----------

	fname: str
		String defining the prefix of the file name and the containing folder.

	data : ndarray
		The 3D stack to write.

	axis : int, optional
		The axis along which the stacking is performed.

	start : int, optional
		Index used for saving the first image.

	croi : tuple, optional
		Tuple defining the indexes range for each axis. It must be follow this notation:
		( (row_start, row_end, row_step),  (col_start, col_end, col_step) )

	digit : int, optional
		Number of digits used for the numbering of the images.

	dtype : dtype, optional
		Data type of the images to save.

	overwrite: bool, optional
		If ``True``, overwrites the output file if it exists. Raises an ``IOError``
		if ``False`` and the output file exists. Default is ``False``.

	"""
	## check dimension data ==3
	if (data.ndim != 3):
		raise ValueError('Array must have 3 dimensions.')

	folder = os.path.dirname(fname)

	if folder == '':
		raise ValueError('File path not valid. Please specify the file path giving\
					at least one folder, e.g.: folder/file.fits ')
	else:
		if not (os.path.isdir(folder)):
			raise ValueError('Folder not exist. %s' %folder)

	if not dtype:
		dtype = data.dtype


	if(axis!=0):
		data  = np.swapaxes(data, 0, axis)
		if(croi):
			newcroi_list       = [[None]*3]*3
			newcroi_list[axis] = croi[0]
			newcroi_list[0]    = croi[axis]
			newcroi_list[3-axis] = croi[3-axis]
			newcroi              = tuple(newcroi_list)

	if not croi:
		slices, rows, cols = data.shape
		newcroi = ((0, slices, 1), (0, rows, 1), (0, cols, 1))
	else:
		if(axis ==0):
			newcroi = croi

	rmin, rmax, rstep = newcroi[1]
	cmin, cmax, cstep = newcroi[2]
	smin, smax, sstep = newcroi[0]

	auslist = list(range(0, data.shape[0], 1))

	subset = auslist[smin:smax:sstep]

	logs.debug('Shape of the re-stacked array %s',  newcroi)

	print('> Saving stack of images:', fname + '_' + '#'*digit + '.fits')
	for iz, s in enumerate(tqdm(subset, unit=' images'), start):
		outfile = fname + '_' + str(iz).zfill(digit) + '.fits'
		write_fits(outfile, data[s, rmin:rmax:rstep, cmin:cmax:cstep].astype(dtype), overwrite)



from tkinter import *
from tkinter import filedialog
import os
import logging
import tkinter

__author__  = "Davide Micieli"
__all__     = ['get_image_gui',
	       'save_filename_gui',
		'get_folder_gui',
		'get_filename_gui',
		'get_screen_resolution']

#logging.basicConfig(level=logging.WARNING)
logs = logging.getLogger(__name__)


def get_image_gui(initialdir='', message='Select image...'):
	"""
	This function opens a dialog box to select an image (TIFF or FITS) and to get
	its file path.

	Parameters
	----------
	initialdir : str, optional
		String defining the path of the initial folder to open in the dialog box.

	message : str, optional
		String defining the dialog box title.

	Returns
	-------
	fname : str
		String defining the file path selected using the dialog box.
	"""

	if not (initialdir):
		initialdir = os.path.abspath(os.sep)

	root = Tk()
	root.withdraw()
	root.wm_attributes('-topmost', 1)

	while True:
		fname = filedialog.askopenfilename(initialdir=initialdir, title=message,
						filetypes = (("Image files", "*.tif *.tiff *.fits"),
						("TIFF files","*.tif *.tiff"),("FITS files","*.fits")))

		if(fname):
			break

	return fname


def save_filename_gui(initialdir='', message='Select folder and the name of the file to save...' ):
	"""
	This function opens a dialog box to select a file to save and get its file path.

	Parameters
	----------
	initialdir : str, optional
		String defining the path of the initial folder to open in the dialog box.

	message : str, optional
		String defining the dialog box title.

	Returns
	-------
	fname : str
		String defining the file path selected using the dialog box.
	"""

	if not (initialdir):
		initialdir = os.path.abspath(os.sep)

	root = Tk()
	root.withdraw()
	root.wm_attributes('-topmost', 1)

	while True:
		fname =  filedialog.asksaveasfilename(initialdir = initialdir, title = message)

		if(fname):
			break


	return fname


def get_folder_gui(initialdir='', message='Select folder...'):
	"""
	This function opens a dialog box to select a folder and get its file path.

	Parameters
	----------
	initialdir : str, optional
		String defining the path of the initial folder to open in the dialog box.

	message : str, optional
		String defining the dialog box title.

	Returns
	-------
	fname : str
		String defining the folder path selected using the dialog box.
	"""

	if not (initialdir):
		initialdir = os.path.abspath(os.sep)

	root = Tk()
	root.withdraw()
	root.wm_attributes('-topmost', 1)


	while True:
		fname = filedialog.askdirectory(initialdir=initialdir, title=message)

		if(fname):
			break

	return fname


def get_filename_gui(initialdir='', message='Select file...', ext=None ):
	"""
	This function opens a dialog box to select a file and get its file path.

	Parameters
	----------
	initialdir : str, optional
		String defining the path of the initial folder to open in the dialog box.

	message : str, optional
		String defining the dialog box title.

	ext : tuple, optional
		Tuple defining the file types to show. It includes the description and a
		shell-style wildcards defining the extension of the files.
		E.g. to filter TIFF images: (('Tiff iamges', '*.tiff *.tif'))

	Returns
	-------
	fname : str
		String defining the file path selected using the dialog box.
	"""

	if not (initialdir):
		initialdir = os.path.abspath(os.sep)

	root = Tk()
	root.withdraw()
	root.wm_attributes('-topmost', 1)

	while True:

		if ext is None:
			fname = filedialog.askopenfilename(initialdir=initialdir, title=message)
		else:
			ext = ( ext, ("All files", "*.*") )
			fname = filedialog.askopenfilename(initialdir=initialdir, title=message,
			 				   filetypes = ext)

		if(fname):
			break

	return fname


def get_screen_resolution():
	"""
	This function returns the screen resolution as tuple.

	Example
	-------
	>>> width, height = ntp.get_screen_resolution()
	"""
	root = tkinter.Tk()
	root.withdraw()
	width = root.winfo_screenwidth()
	height = root.winfo_screenheight()
	return (width, height)



#-----------------------------------------------------------------------
#Copyright 2013 Centrum Wiskunde & Informatica, Amsterdam
#
#Author: Daniel M. Pelt
#Contact: D.M.Pelt@cwi.nl
#Website: http://dmpelt.github.io/pynnfbp/
#
#
#This file is part of the PyNN-FBP, a Python implementation of the
#NN-FBP tomographic reconstruction method.
#
#PyNN-FBP is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.
#
#PyNN-FBP is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with PyNN-FBP. If not, see <http://www.gnu.org/licenses/>.
#
#-----------------------------------------------------------------------

from .astra_plugin import plugin_prepare, plugin_train, plugin_rec



import numpy as np
import numexpr as ne
import matplotlib.pyplot as plt
from skimage.measure import profile_line, compare_ssim
from numpy.linalg import norm
from read_roi import read_roi_file
from neutompy.postproc.crop import *
from neutompy.image.image import get_rect_coordinates_from_roi
import logging
import os
from scipy.optimize import curve_fit
from scipy.special import erf
from scipy import signal

logs = logging.getLogger(__name__)

__author__  = "Davide Micieli"
__all__     = ['CNR', 'NRMSE', 'SSIM', 'FWHM', 'get_line_profile', 'GMSD']


def CNR(img, croi_signal=[], croi_background=[], froi_signal=[], froi_background=[]):
	"""
	This function computes the Contrast-to-Noise Ratio (CNR) as reported in
	the equation (2.7) of [1]_.
	The ROI of the signal and the background can be defined using two lists of
	coordinates or two ImageJ .roi files.

	Parameters
	----------
	img :  2d array
		The array representing the image.

	croi_signal : list
		List that contains the following coordinate of the signal roi: [rowmin, rowmax, colmin, colmax].

	croi_background : list
		List that contains the following coordinate of the background roi: [rowmin, rowmax, colmin, colmax].

	froi_signal : string
		Path of the imagej file containing the rectangular ROI of the signal.

	froi_background : string
		Path of the imagej file containing the rectangular ROI of the background.

	Returns
	-------
	CNR : float
		The CNR value computed using the ROIs given.

	References
	----------
	.. [1] D. Micieli et al., A comparative analysis of reconstruction methods
	 applied to Neutron Tomography, Journal of Instrumentation, Volume 13,
	 June 2018.
	"""

	if(img.ndim != 2):
		raise ValueError("The input array must have 2 dimensions.")

	if(croi_signal and froi_signal):
		raise ValueError("Only one method to define the ROI is accepted. Please pass croi_singal or froi_signal.")

	if(croi_background and froi_background):
		raise ValueError("Only one method to define the ROI is accepted. Please pass croi_background or froi_background.")

	if(croi_signal):
		rowmin, rowmax, colmin, colmax = croi_signal
	if(froi_signal):
		rowmin, rowmax, colmin, colmax = get_rect_coordinates_from_roi(froi_signal)

	signal = img[rowmin:(rowmax+1),  colmin:(colmax+1)]

	if(croi_background):
		rowmin, rowmax, colmin, colmax = croi_background
	elif(froi_background):
		rowmin, rowmax, colmin, colmax = get_rect_coordinates_from_roi(froi_background)

	background = img[rowmin:(rowmax+1),  colmin:(colmax+1)]

	cnr_val = ( signal.mean() - background.mean() ) / background.std()

	return cnr_val




def get_line_profile(image, start=(), end=(), froi='', ShowPlot=True, PlotTitle='Profile', linewidth=1, order=1, mode='constant', cval=0.0):
	"""
	This function returns the intensity profile of an image measured along a line defined by the points:

			start = (x_start, y_start)  [i.e. (col_start row_start)]
			end   = (x_end, y_en)       [i.e. (col_end row_end)]

	or an ImageJ .roi file containing the line selection. A plot representing the intensity profile can be shown.

    Parameters
    ----------
    image : ndarray
        The image grayscale (2D array) or a stack of images (3d array) with shape (slices, rows, columns).
        Ffor a 3D array the first axis represents the image index.

    start : 2-tuple of numeric scalar (float or int) (x y) [i.e. (col row)]
        The start point of the scan line.

    end : 2-tuple of numeric scalar (float or int) (x y) [i.e. (col row)]
        The end point of the scan line. The destination point is *included*
        in the profile, in constrast to standard numpy indexing.

    froi : string
        Path of the imagej file containing the line selection.

    ShowPlot : bool
        If True a canvas is created representing the Plot Profile.

    linewidth : int, optional
        Width of the scan, perpendicular to the line

    order : int in {0, 1, 2, 3, 4, 5}, optional
        The order of the spline interpolation to compute image values at
        non-integer coordinates. 0 means nearest-neighbor interpolation.

    mode : {'constant', 'nearest', 'reflect', 'mirror', 'wrap'}, optional
        How to compute any values falling outside of the image.

    cval : float, optional
        If `mode` is 'constant', what constant value to use outside the image.

    Returns
    -------
    return_value : array
        The intensity profile along the scan line. The length of the profile
        is the ceil of the computed length of the scan line.
	"""

	if(froi and (start or end)):
		raise ValueError('Please indicate an ImageJ .roi file or two tuples of coordinates.')

	if(froi):
		# read roi file
		roi = read_roi_file(froi)

		# convert keys to list
		l = list(roi.keys())

		# extract type
		tp = roi[l[0]]['type']

		if(tp=='line'):
			start = np.floor([roi[l[0]]['x1'], roi[l[0]]['y1']])
			end   = np.floor([roi[l[0]]['x2'], roi[l[0]]['y2']])
		else:
			raise ValueError("File specified does not contain a line selection.")


	start = start[::-1]
	end   = end[::-1]

	if(image.ndim == 3):
		image = image.swapaxes(0,2).swapaxes(0,1)

	y = profile_line(image, start, end, linewidth, order, mode, cval)


	if(ShowPlot):
		plt.figure(PlotTitle);
		plt.plot(y, '-');
		plt.show(block=False)

	if(image.ndim == 3):
		y = y.swapaxes(0,1)

	return y



def NRMSE(img, ref,  mask='whole'):
	"""
	This function computes the Normalized Root Mean Square Error (see eq. 2.9
	in [1]_) of an image respect to a reference image.

	Parameters
	----------
	img: 2d array
		Test image

	ref: 2d array
		Reference image

	mask:
		It represents the type of the ROI where the NRMSE is computed.
		You can specify: 'whole'  -> full image
						 'circ'   -> circular mask
						 'custom' -> custom mask, defined as a boolean matrix of the same shape of the images (img and ref)
	mask: 2d bool array
		Curstom mask of the same shape of the images, where the NRMSE is computed.

	Returns
	-------
	NRMSE: float
		The NRMSE value computed within the roi specified.
	"""

	if(img.shape != ref.shape):
		raise ValueError('The two input arrays must have the same shape.')

	if(img.ndim != 2 or ref.ndim != 2 ):
		raise ValueError('The two input arrays must be 2D.')

	if (type(mask) is str):

		if(mask.endswith('.roi')):

			if not os.path.isfile(mask):
				raise ValueError("The specified file .roi does not exixst.")

			rowmin, rowmax, colmin, colmax = get_rect_coordinates_from_roi(mask)
			vimg = img[rowmin:(rowmax+1),  colmin:(colmax+1)]
			vref = ref[rowmin:(rowmax+1),  colmin:(colmax+1)]

		elif((mask == 'whole') or (mask is None)):

			vimg = img
			vref = ref

		elif(mask == 'circ'):
			nrow, ncol = img.shape
			mask = get_circular_mask(nrow, ncol)
			vimg = img[mask]
			vref = ref[mask]
		else:
			raise ValueError('The mask variable must be a string representing a file path or the keywords `whole`, `circ`.')

	elif(type(mask) is np.ndarray):

		if(mask.dtype is not np.dtype('bool')):
			raise ValueError('The mask must be a boolean matrix.')

		vimg = img[mask]
		vref = ref[mask]

	else:
		raise ValueError('The type of the region-of-interest for the NRMSE computation is not valid')

	vimg = vimg.astype(np.float64)
	vref = vref.astype(np.float64)

	val = norm(vimg - vref) / norm(vref)

	return val


def sigmoid_gaus(x, k0, k1, k2, k3):
	"""
	This function computes a general formula of the Gauss error function.
	The exact formula implemented is f(x) = 0.5 * k0 * (Erf(k1*(x - k2)) + 1.0) + k3,
	where Erf is the  Gauss error function, k0, k1, k2, k3 and k4 are constants.

	Parameters
	----------
	x : 1d array
		The vector containing the abscissa values.

	k0 : float
		The height of the sigmoid function.

	k1 : float
		This parameters set the steepness of the sigmoid.

	k2 : float
		The midpoint of the sigmoid

	k3 : float
		The bias parameter.

	Returns
	-------
	val : float
		The function value.
	"""
	val = 0.5 * k0 * (erf(k1*(x - k2)) + 1.0) + k3
	return val



def FWHM(profile, yerr=None):
	"""
	This functions computes an edge quality metric from a profile of a sharp edge.
	The profile is fitted with a generic Gauss sigmoid function.
	The fitting function was then differentiated and the FWHM of the gaussian
	obtained is returned by the function.
	This method is described in detail in [1]_.

	Parameters
	----------
	profile : 1d array
		The line profile of the sharp edge.

	yerr : 1d array, optional
		The vector containing the standard deviation of the edge profile.
		If not specified the standard deviation of each point is assumed equal
		to 1.0.

	Returns
	-------
	fwhm : float
		The FWHM mean value.

	fwhm_err : float
		The FWHM error value

	profile_fitted: 1d array
		The Gauss sigmoid function evaluated for the fitting parameters.

	popt : list
		List containing the fitting parameters.

	perr : list
		List containing the errors of the fitting parameters.

	"""

	npoints = profile.size
	xdata   = np.arange(npoints)
	pin    = [None]*4
	pin[0] = profile.max() - profile.min()
	pin[1] = 0.5
	pin[2] = npoints/2
	pin[3] = profile.min()


	popt, pcov = curve_fit(sigmoid_gaus, xdata, profile, p0=pin, sigma=yerr)
	perr = np.sqrt(np.diag(pcov))

	p1     = popt[1]
	p1_err = perr[1]

	k = 2.0 * np.sqrt( np.log(2) )

	fwhm     = k / p1
	fwhm_err = k*p1_err / (p1**2)

	profile_fitted = sigmoid_gaus(xdata, *popt)

	return fwhm, fwhm_err, profile_fitted, popt, perr


def SSIM(img1, img2, circ_crop=True, L=None, K1=0.01, K2 = 0.03, sigma=1.5, local_ssim=False):
	"""
	This function computes the Structural Similarity Index (SSIM) [2]_.
	It returns global SSIM value and, optionally, the local SSIM map.

	Parameters
	----------
	img1: 2d array
		The first image to compare. SSIM index satisfies the condition of simmetry: SSIM(img1, img2) = SSIM(img2, img1)

	img2: 2d array
		The second image to compare. SSIM index satisfies the condition of simmetry: SSIM(img1, img2) = SSIM(img2, img1)

	circular_crop : bool, optional
		If True (default) the images are cropped with a circular mask,
		otherwise the SSIM is computed over the entire image.

	L : float, optional
		The data range of the input images (distance between minimum and
		maximum possible values). By default, this is estimated from
		the image data-type.

	K1 : float, optional
		A constant that prevents the division by zero (see [2]_).

	K2 : float, optional
		A constant that prevents the division by zero (see [2]_).

	sigma : float, optional
		The standard deviation of the Gaussian filter. This parameter
		sets the minimum scale at which the quality is evaluated.

	local_ssim: float, optional
		If True, the function returns the local SSIM map.

	Returns
	-------
	ssim : float
		The global SSIM index.

	map : 2d array
		The bidimendional map of the local SSIM index. This is only returned
		if `local_ssim` is set to True.

	References
	----------
	.. [2] Wang, Z., Bovik, A. C., Sheikh, H. R., & Simoncelli, E. P.
		(2004). Image quality assessment: From error visibility to
		structural similarity. IEEE Transactions on Image Processing,
		13, 600-612.
		https://ece.uwaterloo.ca/~z70wang/publications/ssim.pdf
	"""

	if(img1.shape != img2.shape):
		raise ValueError('The input images must have the same shape.')

	vimg1 = np.zeros(img1.shape)
	vimg2 = np.zeros(img2.shape)

	if(circ_crop):
		nrow, ncol = img1.shape
		mask = get_circular_mask(nrow, ncol)
		vimg1[mask] = img1[mask]
		vimg2[mask] = img2[mask]
	else:
		vimg1 = img1
		vimg2 = img2

	val = compare_ssim(vimg1, vimg2, data_range=L, gaussian_weights=True, sigma=sigma, k1=K1, K2=K2, use_sample_covariance=False, full=local_ssim)
	return val


def GMSD(img, ref, rescale=True, map=False):
	"""
	This function computes the Gradient Magnitude Similarity Deviation (GMSD).
	This is a Python version of the Matlab script provided by the authors in [3]_

	Parameters
	----------
	img : 2d array
		Image to compare.

	ref : 2d array
		Reference image.

	rescale : bool, optional
		If True the input images were rescaled in such a way that `ref` has a
		maximum pixel value of 255. If False no rescaling is performed.
		Default value is True.

	map : bool, optional
		If True the GMSD and GMS map are returned in a tuple.
		Default value is False.

	Returns
	-------
	gmsd_val : float
		The GMSD value.

	gms_map : 2d array
		The GMS map, returned only if `map` is True.

	References
	----------
	.. [3] Wufeng Xue, Lei Zhang, Xuanqin Mou, and Alan C. Bovik,  "Gradient Magnitude
		Similarity Deviation: A Highly Efficient Perceptual Image Quality Index",
		http://www.comp.polyu.edu.hk/~cslzhang/IQA/GMSD/GMSD.htm
	"""
	if rescale:
		k = 255.0 / ref.max()
	else:
		k = 1.0

	T = 170
	downscale = 2
	hx = (1.0/3.0) * np.array([[1, 0, -1],
							   [1, 0, -1],
							   [1, 0, -1]])
	hy = hx.T

	avg_kernel = np.ones((2, 2)) / 4.0
	avg_ref = signal.convolve2d(k * ref, avg_kernel, mode='same', boundary='symm')[::downscale,::downscale]
	avg_img = signal.convolve2d(k * img, avg_kernel, mode='same', boundary='symm')[::downscale,::downscale]

	ref_dx =  signal.convolve2d(avg_ref, hx, mode='same', boundary='symm')
	ref_dy =  signal.convolve2d(avg_ref, hy, mode='same', boundary='symm')
	MG_ref =  np.sqrt(ref_dx**2 + ref_dy**2)

	img_dx =  signal.convolve2d(avg_img, hx, mode='same', boundary='symm')
	img_dy =  signal.convolve2d(avg_img, hy, mode='same', boundary='symm')
	MG_img =  np.sqrt(img_dx**2 + img_dy**2)

	gms_map = (2*MG_ref*MG_img + T) / (MG_img**2 + MG_ref**2 + T)

	gmsd_val = np.std(gms_map)

	if map:
		return gmsd_val, gms_map
	else:
		return gmsd_val



# -----------------------------------------------------------------------
# Copyright: 2010-2018, imec Vision Lab, University of Antwerp
#            2013-2018, CWI, Amsterdam
#
# Contact: astra@astra-toolbox.com
# Website: http://www.astra-toolbox.com/
#
# This file is part of the ASTRA Toolbox.
#
#
# The ASTRA Toolbox is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The ASTRA Toolbox is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with the ASTRA Toolbox. If not, see <http://www.gnu.org/licenses/>.
#
# -----------------------------------------------------------------------

from astra import data2d
from astra import data3d
from astra import projector
from astra import projector3d
from astra import creators
from astra import algorithm
from astra import functions
import numpy as np
from six.moves import reduce
try:
    from six.moves import range
except ImportError:
    # six 1.3.0
    from six.moves import xrange as range

import operator
import scipy.sparse.linalg

class OpTomo(scipy.sparse.linalg.LinearOperator):
    """Object that imitates a projection matrix with a given projector.

    This object can do forward projection by using the ``*`` operator::

        W = astra.OpTomo(proj_id)
        fp = W*image
        bp = W.T*sinogram

    It can also be used in minimization methods of the :mod:`scipy.sparse.linalg` module::

        W = astra.OpTomo(proj_id)
        output = scipy.sparse.linalg.lsqr(W,sinogram)

    :param proj_id: ID to a projector.
    :type proj_id: :class:`int`
    """

    def __init__(self,proj_id):
        self.dtype = np.float32
        try:
            self.vg = projector.volume_geometry(proj_id)
            self.pg = projector.projection_geometry(proj_id)
            self.data_mod = data2d
            self.appendString = ""
            if projector.is_cuda(proj_id):
                self.appendString += "_CUDA"
        except Exception:
            self.vg = projector3d.volume_geometry(proj_id)
            self.pg = projector3d.projection_geometry(proj_id)
            self.data_mod = data3d
            self.appendString = "3D"
            if projector3d.is_cuda(proj_id):
                self.appendString += "_CUDA"

        self.vshape = functions.geom_size(self.vg)
        self.vsize = reduce(operator.mul,self.vshape)
        self.sshape = functions.geom_size(self.pg)
        self.ssize = reduce(operator.mul,self.sshape)

        self.shape = (self.ssize, self.vsize)

        self.proj_id = proj_id

        self.transposeOpTomo = OpTomoTranspose(self)
        try:
            self.T = self.transposeOpTomo
        except AttributeError:
            # Scipy >= 0.16 defines self.T using self._transpose()
            pass

    def _transpose(self):
        return self.transposeOpTomo

    def __checkArray(self, arr, shp):
        if len(arr.shape)==1:
            arr = arr.reshape(shp)
        if arr.dtype != np.float32:
            arr = arr.astype(np.float32)
        if arr.flags['C_CONTIGUOUS']==False:
            arr = np.ascontiguousarray(arr)
        return arr

    def _matvec(self,v):
        """Implements the forward operator.

        :param v: Volume to forward project.
        :type v: :class:`numpy.ndarray`
        """
        return self.FP(v, out=None).ravel()

    def rmatvec(self,s):
        """Implements the transpose operator.

        :param s: The projection data.
        :type s: :class:`numpy.ndarray`
        """
        return self.BP(s, out=None).ravel()

    def __mul__(self,v):
        """Provides easy forward operator by *.

        :param v: Volume to forward project.
        :type v: :class:`numpy.ndarray`
        """
        # Catch the case of a forward projection of a 2D/3D image
        if isinstance(v, np.ndarray) and v.shape==self.vshape:
            return self._matvec(v)
        return scipy.sparse.linalg.LinearOperator.__mul__(self, v)

    def reconstruct(self, method, s, iterations=1, extraOptions = None):
        """Reconstruct an object.

        :param method: Method to use for reconstruction.
        :type method: :class:`string`
        :param s: The projection data.
        :type s: :class:`numpy.ndarray`
        :param iterations: Number of iterations to use.
        :type iterations: :class:`int`
        :param extraOptions: Extra options to use during reconstruction (i.e. for cfg['option']).
        :type extraOptions: :class:`dict`
        """
        if extraOptions == {}:
            opts={}
        
        if extraOptions is None:
            opts={}
		  
        s = self.__checkArray(s, self.sshape)
        sid = self.data_mod.link('-sino',self.pg,s)
        v = np.zeros(self.vshape,dtype=np.float32)
        vid = self.data_mod.link('-vol',self.vg,v)
        cfg = creators.astra_dict(method)
        cfg['ProjectionDataId'] = sid
        cfg['ReconstructionDataId'] = vid
        cfg['ProjectorId'] = self.proj_id
        if 'FilterType' in list(extraOptions.keys()):
            cfg['FilterType'] = extraOptions['FilterType']
            opts = {key: extraOptions[key] for key in extraOptions if key != 'FilterType'}
        else:
            opts = extraOptions
        
        cfg['option'] = opts
        alg_id = algorithm.create(cfg)
        algorithm.run(alg_id,iterations)
        algorithm.delete(alg_id)
        self.data_mod.delete([vid,sid])
        return v

    def FP(self,v,out=None):
        """Perform forward projection.

        Output must have the right 2D/3D shape. Input may also be flattened.

        Output must also be contiguous and float32. This isn't required for the
        input, but it is more efficient if it is.

        :param v: Volume to forward project.
        :type v: :class:`numpy.ndarray`
        :param out: Array to store result in.
        :type out: :class:`numpy.ndarray`
        """

        v = self.__checkArray(v, self.vshape)
        vid = self.data_mod.link('-vol',self.vg,v)
        if out is None:
            out = np.zeros(self.sshape,dtype=np.float32)
        sid = self.data_mod.link('-sino',self.pg,out)

        cfg = creators.astra_dict('FP'+self.appendString)
        cfg['ProjectionDataId'] = sid
        cfg['VolumeDataId'] = vid
        cfg['ProjectorId'] = self.proj_id
        fp_id = algorithm.create(cfg)
        algorithm.run(fp_id)

        algorithm.delete(fp_id)
        self.data_mod.delete([vid,sid])
        return out

    def BP(self,s,out=None):
        """Perform backprojection.

        Output must have the right 2D/3D shape. Input may also be flattened.

        Output must also be contiguous and float32. This isn't required for the
        input, but it is more efficient if it is.

        :param : The projection data.
        :type s: :class:`numpy.ndarray`
        :param out: Array to store result in.
        :type out: :class:`numpy.ndarray`
        """
        s = self.__checkArray(s, self.sshape)
        sid = self.data_mod.link('-sino',self.pg,s)
        if out is None:
            out = np.zeros(self.vshape,dtype=np.float32)
        vid = self.data_mod.link('-vol',self.vg,out)

        cfg = creators.astra_dict('BP'+self.appendString)
        cfg['ProjectionDataId'] = sid
        cfg['ReconstructionDataId'] = vid
        cfg['ProjectorId'] = self.proj_id
        bp_id = algorithm.create(cfg)
        algorithm.run(bp_id)

        algorithm.delete(bp_id)
        self.data_mod.delete([vid,sid])
        return out




class OpTomoTranspose(scipy.sparse.linalg.LinearOperator):
    """This object provides the transpose operation (``.T``) of the OpTomo object.

    Do not use directly, since it can be accessed as member ``.T`` of
    an :class:`OpTomo` object.
    """
    def __init__(self,parent):
        self.parent = parent
        self.dtype = np.float32
        self.shape = (parent.shape[1], parent.shape[0])
        try:
            self.T = self.parent
        except AttributeError:
            # Scipy >= 0.16 defines self.T using self._transpose()
            pass

    def _matvec(self, s):
        return self.parent.rmatvec(s)

    def rmatvec(self, v):
        return self.parent.matvec(v)

    def _transpose(self):
        return self.parent

    def __mul__(self,s):
        # Catch the case of a backprojection of 2D/3D data
        if isinstance(s, np.ndarray) and s.shape==self.parent.sshape:
            return self._matvec(s)
        return scipy.sparse.linalg.LinearOperator.__mul__(self, s)



#__version__ = '1.0.0'
#execfile('version.py')


#version = {}
#with open("neutompy/version.py") as fp:
#    exec(fp.read(), version)
#__version__ = version['__version__']

import os

# load __version__ variable
package_root_dir = os.path.dirname(os.path.abspath(__file__))
file_version     = os.path.join(package_root_dir, "version.py")
exec(open(file_version).read())


from .image.image import *
from .image.rebin import *
from .preproc.preproc import *
from .recon.recon import *
from .misc.uitools import *
from .postproc.convert import *
from .postproc.crop import *
from .metrics.metrics import *



#-----------------------------------------------------------------------
#Copyright 2015 Daniel M. Pelt
#
#Contact: D.M.Pelt@cwi.nl
#Website: http://www.dmpelt.com
#
#
#This file is part of the PyNN-FBP, a Python implementation of the
#NN-FBP tomographic reconstruction method.
#
#PyNN-FBP is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.
#
#PyNN-FBP is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with PyNN-FBP. If not, see <http://www.gnu.org/licenses/>.
#
#-----------------------------------------------------------------------
import astra
import numpy as np
from six.moves import range
from scipy.signal import fftconvolve
import scipy.ndimage.filters as snf
import numpy.linalg as na
import glob
import tifffile
import scipy.io as sio

import os, errno

import random

from .TrainingData import MATTrainingData
from neutompy.image.image import read_image

def mkdir_p(path):
    try:
        os.makedirs(path)
    except OSError as exc: # Python >2.5
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else: raise

def customFBP(W, f, s):
    sf = np.zeros_like(s)
    padded = np.zeros(s.shape[1]*2)
    l = int(s.shape[1]/2.)
    r = l+s.shape[1]
    bl = f.shape[0]/len(s)
    if len(f.shape)==1:
        f = np.tile(f,(sf.shape[0],1))
    for i in range(sf.shape[0]):
        padded[l:r] = s[i]
        padded[:l] = padded[l]
        padded[r:] = padded[r-1]
        sf[i] = fftconvolve(padded,f[i],'same')[l:r]
    return (W.T*sf).reshape(W.vshape)


class plugin_prepare(astra.plugin.ReconstructionAlgorithm2D):
    """Prepares a training set for the NN-FBP method [1].

    Options:

    'hqrecfiles': HQ reconstruction files (in TIFF format), e.g. '/path/to/rec*.tiff'.
    'traindir': folder where output training files should be stored
    'z_id': z-index of current slice
    'npick' (optional): number of random pixels to pick (per slice)
    'nlinear' (optional): number of linear steps in exponential binning
    'extra_ids' (optional): extra ASTRA data ids to use during reconstruction
    'angle_dependent' (optional): compute an angle-dependent filter if set to "True"

    [1] Pelt, D. M., & Batenburg, K. J. (2013). Fast tomographic reconstruction
        from limited data using artificial neural networks. Image Processing,
        IEEE Transactions on, 22(12), 5238-5251.
    """

    astra_name="NN-FBP-prepare"


    def initialize(self, cfg, hqrecfiles, z_id, traindir, nlinear=2, npick=100, extra_ids=None, angle_dependent=False):
        if extra_ids==None:
            extra_ids = []
        self.extra_s = [astra.data2d.get_shared(i) for i in extra_ids]
        self.W = astra.OpTomo(cfg['ProjectorId'])
        self.npick = npick
        self.td = traindir
        mkdir_p(traindir)
        # self.rec = tifffile.imread(sorted(glob.glob(hqrecfiles))[z_id])
        self.rec = read_image(sorted(glob.glob(hqrecfiles))[z_id])
        self.bck = (self.W.T*np.ones_like(self.s)<self.s.shape[0]-0.5).reshape(self.v.shape)
        self.outfn = traindir + os.sep + "train_{:05d}.mat".format(z_id)
        self.z_id = z_id
        fs = self.s.shape[1]
        if fs%2==0:
            fs += 1
        mf = int(fs/2)
        w=1
        c=mf
        bas = np.zeros(fs,dtype=np.float32)
        self.basis = []
        count=0
        while c<fs:
            bas[:]=0
            l = c
            r = c+w
            if r>fs: r=fs
            bas[l:r]=1
            if l!=0:
                l = fs-c-w
                r = l+w
                if l<0: l=0
                bas[l:r]=1
            self.basis.append(bas.copy())
            c += w
            count += 1
            if count>nlinear:
                w=2*w
        if angle_dependent=="True":
            basis_ang = []
            bas_ang = np.zeros((self.s.shape[0],fs),dtype=np.float32)
            for bas in self.basis:
                for i in range(self.s.shape[0]):
                    bas_ang[i] = bas
                    basis_ang.append(bas_ang.copy())
                    bas_ang[i][:] = 0
            self.basis = basis_ang
        self.nf = len(self.basis)

    def run(self, iterations):
        out = np.zeros((self.npick,(1+len(self.extra_s))*self.nf+1))
        pl = np.random.random((self.npick,2))
        pl[:,0]*=self.v.shape[0]
        pl[:,1]*=self.v.shape[1]
        pl = pl.astype(np.int)
        for i in range(self.npick):
            while self.bck[pl[i,0],pl[i,1]]:
                pl[i,0] = np.random.random(1)*self.v.shape[0]
                pl[i,1] = np.random.random(1)*self.v.shape[1]
        out[:,-1] = self.rec[pl[:,0],pl[:,1]]
        for i, bas in enumerate(self.basis):
            img = customFBP(self.W, bas, self.s)
            out[:,i] = img[pl[:,0],pl[:,1]]
            for l, s in enumerate(self.extra_s):
                img = customFBP(self.W, bas, s)
                out[:,i+(l+1)*self.nf] = img[pl[:,0],pl[:,1]]
        sio.savemat(self.outfn,{'mat':out},do_compression=True)
        #astra.log.info('Slice {} done...'.format(self.z_id))

import numexpr
import time
import scipy.sparse as ss
import scipy.linalg as la
try:
    import scipy.linalg.fblas as fblas
    hasfblas=True
except:
    hasfblas=False

def sigmoid(x):
    '''Sigmoid function'''
    return numexpr.evaluate("1./(1.+exp(-x))")

class Network(object):
    '''
    The neural network object that performs all training and reconstruction.

    :param nHiddenNodes: The number of hidden nodes in the network.
    :type nHiddenNodes: :class:`int`
    :param projector: The projector to use.
    :type projector: A ``Projector`` object (see, for example: :mod:`nnfbp.SimpleCPUProjector`)
    :param trainData: The training data set.
    :type trainData: A ``DataSet`` object (see: :mod:`nnfbp.DataSet`)
    :param valData: The validation data set.
    :type valData: A ``DataSet`` object (see: :mod:`nnfbp.DataSet`)
    :param reductor: Optional reductor to use.
    :type reductor: A ``Reductor`` object (see: :mod:`nnfbp.Reductors`, default:``LogSymReductor``)
    :param nTrain: Number of pixels to pick out of training set.
    :type nTrain: :class:`int`
    :param nVal: Number of pixels to pick out of validation set.
    :type nVal: :class:`int`
    :param tmpDir: Optional temporary directory to use.
    :type tmpDir: :class:`string`
    :param createEmptyClass: Used internally when loading from disk, to create an empty object. Do not use directly.
    :type createEmptyClass: :class:`boolean`
    '''

    def __init__(self, nHiddenNodes, trainData, valData, setinit=None):
        self.tTD = trainData
        self.vTD = valData
        self.nHid = nHiddenNodes
        self.nIn = self.tTD.getDataBlock(0).shape[1]-1
        self.jacDiff = np.zeros((self.nHid) * (self.nIn+1) + self.nHid + 1);
        self.jac2 = np.zeros(((self.nHid) * (self.nIn+1) + self.nHid + 1, (self.nHid) * (self.nIn+1) + self.nHid + 1))
        self.setinit = setinit

    def __inittrain(self):
        '''Initialize training parameters, create actual training and validation
        sets by picking random pixels from the datasets'''
        self.l1 = 2 * np.random.rand(self.nIn+1, self.nHid) - 1
        if self.setinit is not None:
            self.l1.fill(0)
            nd = self.nIn/self.setinit[0]
            for i,j in enumerate(self.setinit[1]):
                self.l1[j*nd:(j+1)*nd,i] = 2 * np.random.rand(nd) - 1
                self.l1[-1,i] = 2 * np.random.rand(1) - 1
        beta = 0.7 * self.nHid ** (1. / (self.nIn))
        l1norm = np.linalg.norm(self.l1)
        self.l1 *= beta / l1norm
        self.l2 = 2 * np.random.rand(self.nHid + 1) - 1
        self.l2 /= np.linalg.norm(self.l2)
        self.minl1 = self.l1.copy()
        self.minl2 = self.l2.copy()
        self.minmax = self.tTD.getMinMax()
        self.tTD.normalizeData(self.minmax[0], self.minmax[1], self.minmax[2], self.minmax[3])
        self.vTD.normalizeData(self.minmax[0], self.minmax[1], self.minmax[2], self.minmax[3])
        self.ident = np.eye((self.nHid) * (self.nIn+1) + self.nHid + 1)


    def __processDataBlock(self,data):
        ''' Returns output values (``vals``), 'correct' output values (``valOut``) and
        hidden node output values (``hiddenOut``) from a block of data.'''
        valOut = data[:, -1].copy()
        data[:, -1] = -np.ones(data.shape[0])
        hiddenOut = np.empty((data.shape[0],self.l1.shape[1]+1))
        hiddenOut[:,0:self.l1.shape[1]] = sigmoid(np.dot(data, self.l1))
        hiddenOut[:,-1] = -1
        rawVals = np.dot(hiddenOut, self.l2)
        vals = sigmoid(rawVals)
        return vals,valOut,hiddenOut



    def __getTSE(self, dat):
        '''Returns the total squared error of a data block'''
        tse = 0.
        for i in range(dat.nBlocks):
            data = dat.getDataBlock(i)
            vals,valOut,hiddenOut = self.__processDataBlock(data)
            tse += numexpr.evaluate('sum((vals - valOut)**2)')
        return tse

    def __setJac2(self):
        '''Calculates :math:`J^T J` and :math:`J^T e` for the training data.
        Used for Levenberg-Marquardt method.'''
        self.jac2.fill(0)
        self.jacDiff.fill(0)
        for i in range(self.tTD.nBlocks):
            data = self.tTD.getDataBlock(i)
            vals,valOut,hiddenOut = self.__processDataBlock(data)
            diffs = numexpr.evaluate('valOut - vals')
            jac = np.empty((data.shape[0], (self.nHid) * (self.nIn+1) + self.nHid + 1))
            d0 = numexpr.evaluate('-vals * (1 - vals)')
            ot = (np.outer(d0, self.l2))
            dj = numexpr.evaluate('hiddenOut * (1 - hiddenOut) * ot')
            I = np.tile(np.arange(data.shape[0]), (self.nHid + 1, 1)).flatten('F')
            J = np.arange(data.shape[0] * (self.nHid + 1))
            Q = ss.csc_matrix((dj.flatten(), np.vstack((J, I))), (data.shape[0] * (self.nHid + 1), data.shape[0]))
            jac[:, 0:self.nHid + 1] = ss.spdiags(d0, 0, data.shape[0], data.shape[0]).dot(hiddenOut)
            Q2 = np.reshape(Q.dot(data), (data.shape[0],(self.nIn+1) * (self.nHid + 1)))
            jac[:, self.nHid + 1:jac.shape[1]] = Q2[:, 0:Q2.shape[1] - (self.nIn+1)]
            if hasfblas:
                self.jac2 += fblas.dgemm(1.0,a=jac.T,b=jac.T,trans_b=True)
                self.jacDiff += fblas.dgemv(1.0,a=jac.T,x=diffs)
            else:
                self.jac2 += np.dot(jac.T,jac)
                self.jacDiff += np.dot(jac.T,diffs)

    def train(self):
        '''Train the network using the Levenberg-Marquardt method.'''
        self.__inittrain()
        mu = 100000.;
        muUpdate = 10;
        prevValError = np.Inf
        bestCounter = 0
        tse = self.__getTSE(self.tTD)
        curTime = time.time()
        self.allls = []
        for i in range(1000000):
            self.__setJac2()
            try:
                dw = -la.cho_solve(la.cho_factor(self.jac2 + mu * self.ident), self.jacDiff)
            except la.LinAlgError:
                break
            done = -1
            while done <= 0:
                self.l2 += dw[0:self.nHid + 1]
                for k in range(self.nHid):
                    start = self.nHid + 1 + k * (self.nIn+1);
                    if self.setinit is not None:
                        nd = self.nIn/self.setinit[0]
                        j = self.setinit[1][k]
                        self.l1[j*nd:(j+1)*nd,k] += dw[start+j*nd:start + (j+1)*nd]
                        self.l1[-1,k] += dw[start+self.nIn]
                    else:
                        self.l1[:, k] += dw[start:start + self.nIn+1]
                newtse = self.__getTSE(self.tTD)
                if newtse < tse:
                    if done == -1:
                        mu /= muUpdate
                    if mu <= 1e-100:
                        mu = 1e-99
                    done = 1;
                else:
                    done = 0;
                    mu *= muUpdate
                    if mu >= 1e20:
                        done = 2
                        break;
                    self.l2 -= dw[0:self.nHid + 1]
                    for k in range(self.nHid):
                        start = self.nHid + 1 + k * (self.nIn+1);
                        if self.setinit is not None:
                            nd = self.nIn/self.setinit[0]
                            j = self.setinit[1][k]
                            self.l1[j*nd:(j+1)*nd,k] -= dw[start+j*nd:start + (j+1)*nd]
                            self.l1[-1,k] -= dw[start+self.nIn]
                        else:
                            self.l1[:, k] -= dw[start:start + self.nIn+1]
                    try:
                        dw = -la.cho_solve(la.cho_factor(self.jac2 + mu * self.ident), self.jacDiff)
                    except la.LinAlgError:
                        done=2
            gradSize = np.linalg.norm(self.jacDiff)
            if done == 2:
                break
            curValErr = self.__getTSE(self.vTD)
            if curValErr > prevValError:
                bestCounter += 1
            else:
                prevValError = curValErr
                self.minl1 = self.l1.copy()
                self.minl2 = self.l2.copy()
                if (newtse / tse < 0.999):
                    bestCounter = 0
                else:
                    bestCounter +=1
            if bestCounter == 50:
                break
            if(gradSize < 1e-8):
                break
            tse = newtse
            astra.log.info('Validation set error: {}'.format(prevValError))
            self.allls.append([self.minl1,self.minl2])
        self.l1 = self.minl1
        self.l2 = self.minl2
        self.valErr = prevValError



    def saveAllToDisk(self,fn):
        for i,k in enumerate(self.allls):
            sio.savemat(fn+"{}.mat".format(i),{'l1':k[0], 'l2':k[1], 'minmax':self.minmax},do_compression=True)

    def saveToDisk(self,fn):
        '''Save a trained network to disk, so that it can be used later
        without retraining.

        :param fn: Filename to save it to.
        :type fn: :class:`string`
        '''
        sio.savemat(fn,{'l1':self.l1,'l2':self.l2,'minmax':self.minmax},do_compression=True)


def plugin_train(traindir, nhid, filter_file, val_rat=0.5, setinit=None, saveAll=False):
    """Traing filters and weights using the NN-FBP method [1].

    Options:

    'traindir': folder where training files are stored
    'nhid': number of hidden nodes to use
    'filter_file': file to store trained filters in
    'val_rat' (optional): fraction of training examples to use as validation
    'saveAll' (optional): save filters at each iteration instead of final filter only

    [1] Pelt, D. M., & Batenburg, K. J. (2013). Fast tomographic reconstruction
        from limited data using artificial neural networks. Image Processing,
        IEEE Transactions on, 22(12), 5238-5251.
    """
    fls = glob.glob(traindir + os.sep + '*.mat')
    random.shuffle(fls)
    nval = int(val_rat*len(fls))
    val = MATTrainingData(fls[:nval])
    trn = MATTrainingData(fls[nval:])
    n = Network(nhid, trn, val,setinit=setinit)
    n.train()
    if saveAll:
        n.saveAllToDisk(filter_file)
    else:
        n.saveToDisk(filter_file)

class plugin_rec(astra.plugin.ReconstructionAlgorithm2D):
    """Reconstructs using the NN-FBP method [1].

    Options:

    'filter_file': file with trained filters to use
    'nlinear' (optional): number of linear steps in exponential binning
    'extra_ids' (optional): extra ASTRA data ids to use during reconstruction
    'angle_dependent' (optional): compute an angle-dependent filter if set to "True"

    [1] Pelt, D. M., & Batenburg, K. J. (2013). Fast tomographic reconstruction
        from limited data using artificial neural networks. Image Processing,
        IEEE Transactions on, 22(12), 5238-5251.
    """

    astra_name="NN-FBP"

    def initialize(self, cfg, filter_file, nlinear=2, extra_ids=None, angle_dependent=False):
        if extra_ids==None:
            extra_ids = []
        self.extra_s = [astra.data2d.get_shared(i) for i in extra_ids]
        self.W = astra.OpTomo(cfg['ProjectorId'])
        fs = self.s.shape[1]
        if fs%2==0:
            fs += 1
        mf = int(fs/2)
        w=1
        c=mf
        bas = np.zeros(fs,dtype=np.float32)
        self.basis = []
        count=0
        while c<fs:
            bas[:]=0
            l = c
            r = c+w
            if r>fs: r=fs
            bas[l:r]=1
            if l!=0:
                l = fs-c-w
                r = l+w
                if l<0: l=0
                bas[l:r]=1
            self.basis.append(bas.copy())
            c += w
            count += 1
            if count>nlinear:
                w=2*w
        if angle_dependent=="True":
            basis_ang = []
            bas_ang = np.zeros((self.s.shape[0],fs),dtype=np.float32)
            for bas in self.basis:
                for i in range(self.s.shape[0]):
                    bas_ang[i] = bas
                    basis_ang.append(bas_ang.copy())
                    bas_ang[i][:] = 0
            self.basis = basis_ang
        self.nf = len(self.basis)
        fl = sio.loadmat(filter_file)
        self.l1 = fl['l1']
        self.l2 = fl['l2'].transpose()
        minmax = fl['minmax'][0]
        minL = minmax[0]
        maxL = minmax[1]
        self.minIn = minmax[2]
        self.maxIn = minmax[3]
        mindivmax = minL/(maxL-minL)
        mindivmax[np.isnan(mindivmax)]=0
        mindivmax[np.isinf(mindivmax)]=0
        divmaxmin = 1./(maxL-minL)
        divmaxmin[np.isnan(divmaxmin)]=0
        divmaxmin[np.isinf(divmaxmin)]=0
        nHid = self.l1.shape[1]
        nsl = len(extra_ids)+1
        dims = [nHid,nsl,]
        dims.extend(self.basis[0].shape)
        self.filters = np.empty(dims)
        self.offsets = np.empty(nHid)
        for i in range(nHid):
            wv = (2*self.l1[0:self.l1.shape[0]-1,i]*divmaxmin).transpose()
            self.filters[i] = np.zeros(dims[1:])
            for t, bas in enumerate(self.basis):
                for l in range(nsl):
                    self.filters[i,l] += wv[t+l*len(self.basis)]*bas
            self.offsets[i] = 2*np.dot(self.l1[0:self.l1.shape[0]-1,i],mindivmax.transpose()) + np.sum(self.l1[:,i])

    def run(self, iterations):
        self.v[:]=0
        for i in range(self.l2.shape[0]-1):
            mult = float(self.l2[i])
            offs = float(self.offsets[i])
            back = customFBP(self.W, self.filters[i,0],self.s)
            for l, s in enumerate(self.extra_s):
                back += customFBP(self.W, self.filters[i,l+1],s)
            self.v[:] += numexpr.evaluate('mult/(1.+exp(-(back-offs)))')
        self.v[:] = sigmoid(self.v-self.l2[-1])
        self.v[:] = (self.v-0.25)*2*(self.maxIn-self.minIn) + self.minIn



import numpy as np
from numpy import sin, cos
from mkl_fft import fft, ifft
from numpy.fft import fftshift, ifftshift
import pywt
import matplotlib.pyplot as plt
from skimage.transform import rotate
from matplotlib.offsetbox import AnchoredText
import numexpr as ne
from read_roi import read_roi_file
import cv2
import SimpleITK as sitk
from neutompy.image.image import get_rect_coordinates_from_roi
from neutompy.misc.uitools import get_screen_resolution
from tqdm import tqdm
import sys
from time import sleep

__author__  = "Davide Micieli"
__all__     = ['draw_ROI',
			   'normalize_proj',
			   'log_transform',
			   'find_COR',
			   'correction_COR',
			   'remove_outliers',
			   'remove_outliers_stack',
			   'remove_stripe',
			   'remove_stripe_stack'
			  ]

def draw_ROI(img, title, ratio=0.85):
	"""
	This function allows to select interactively a rectangular region of interest
	(ROI) over an image. The function returns the ROI coordinates.

	Parameters
	----------
	img : 2d array
		The image on which the dose roi is drawn.

	title : str
		String defining the title of the window shown.

	ratio : float, optional
		The filling ratio of the window respect to the screen resolution.
		It must be a number between 0 and 1. The default value is 0.85.

	Returns
	-------
	rowmin : int
		The minimum row coordinate.

	rowmax : int
		The maximum row coordinate.

	colmin : int
		The minimum column coordinate.

	colmax : int
		The maximum column coordinate.

	"""
	if not (0 < ratio <= 1):
		raise ValueError('The variable ratio must be between 0 and 1.')

	if (img.ndim != 2):
		raise ValueError("The image array must be two-dimensional.")

	# window size settings
	(width, height) = get_screen_resolution()
	scale_width = width / img.shape[1]
	scale_height = height / img.shape[0]
	scale = min(scale_width, scale_height)*ratio
	window_width = int(img.shape[1] * scale)
	window_height = int(img.shape[0] * scale)

	img = img.astype(np.float32)
	mu = np.nanmedian(img.ravel())
	finite_vals = np.nonzero(np.isfinite(img))
	s  = img[finite_vals].std()
	img = img/(mu+2*s)  # normalization can be improved
	imgshow = np.clip(img, 0, 1.0)

	condition = True
	while condition:
		cv2.namedWindow(title, cv2.WINDOW_NORMAL)
		cv2.resizeWindow(title,  window_width, window_height)
		r = cv2.selectROI(title, imgshow, showCrosshair = False, fromCenter = False)

		if (r == (0,0,0,0)):
			condition = True
			print('> ROI cancelled')
		else:
			condition = False
			#print('> ROI selected ')
			rowmin, rowmax, colmin, colmax = r[1], r[1] + r[3], r[0], r[0] + r[2]
			print( 'ROI selected: ymin =', rowmin, ', ymax =', rowmax, ', xmin =', colmin, ', xmax =', colmax)
			cv2.destroyWindow(title)

	return rowmin, rowmax, colmin, colmax




def _normalize (proj, dark, flat, mode='mean', min_denom=1.0e-6, out=None):
	"""
	This function computes the normalization  of the projection data using dark
	and flat images by performing the ratio: (proj - dark) / (flat - dark).
	The ratio is computed using the mean or the median of the dark and flat
	images (defined by the mode). The division by zero is prevented by assigning
	to the denominator a small value (min_denom) if (flat - dark  == 0). The
	result can be stored in an output array by using the 'out' parameter. If
	same as proj, the computation is done in place.
	"""

	if(min_denom<=0.0):
		raise ValueError('The parameter min_ratio must be positive.')

	if (mode == 'mean'):
		func = np.mean
	elif(mode == 'median'):
		func = np.median
	else:
		raise ValueError('Not valid function for projecting flat and dark images\
		 				  along z axis. Set mean or median')

	mean_flat  = func(flat, axis=0).astype(np.float32)
	mean_dark  = func(dark, axis=0).astype(np.float32)

	min_denom = np.float32(min_denom)

	# numerator
	out = np.zeros(proj.shape, dtype=np.float32)
	out = ne.evaluate('proj-mean_dark', out = out)

	# denominator
	den = ne.evaluate('mean_flat-mean_dark')
	den = ne.evaluate('where(den<min_denom, min_denom, den)', out = den)

	ne.evaluate('out/den', out=out, truediv=True)

	return out


def normalize_proj(proj, dark, flat,  proj_180=None, out=None,
						 dose_file='', dose_coor=(), dose_draw=True,
						 crop_file='', crop_coor=(), crop_draw=True,
						 min_denom=1.0e-6,  min_ratio=1e-6, max_ratio=10.0,
						 mode='mean', log=False,  sino_order=False, show_opt='mean'):
	"""
	This function computes the normalization of the projection data using dark
	and flat images. If the source intensity is not stable the images can
	be normalized respect to the radiation dose (see formula 2.2 in [1]_).
	In this case, the user must specify a region (the dose ROI) where the sample
	never appears. If not interested in reconstructing the entire field-of-view,
	the normalization can be performed using only a region of interest (crop ROI)
	of all projections.
	The dose ROI and the crop ROI can be drawn interactively on the image or
	specified using the index coordinates or an ImageJ .roi file. A stack of dark
	and flat images is required and the median or the mean of the stack is
	used to compute the main fomula.

	Parameters
	----------
	proj : ndarray
		A three-dimensional stack of raw projections.
		The 0-axis represents theta.

	dark: ndarray
		A three-dimensional stack of dark-field images.

	flat : ndarray
		A three-dimensional stack of flat-field images.

	proj_180: 2d arrays, optional
		The projection at 180 degree. Specify it only if it is not already
		included in the stack `proj`. It is disabled by default (None).

	out : ndarray, optional
		The output array returned by the function. If it is the same as proj,
		the computation will be done in place.

	dose_file : str, optional
		String defining the path or the name of the Image ROI file that includes
		the dose ROI.

	dose_coor : tuple, optional
		Tuple defining the indexes range for each axis of the ROI dose.
		Specify it in this way: ``(row_start, row_end, col_start, col_end)``

	dose_draw : bool, optional
		If True the dose ROI is selected interactively on the image by the user.
		The default is True.

	crop_file : str, optional
		String defining the path or the name of the Image ROI file that includes
		the ROI to crop.

	crop_coor : tuple, optional
		Tuple defining the indexes range, for each axis, of the ROI to crop.
		Specify it in this way: ``(row_start, row_end, col_start, col_end)``

	crop_draw : bool, optional
		If True the ROI to crop is selected interactively on the image by the user.
		The default is True.

	min_denom : float, optional
		Minimum permitted value of the denominator. It must be a small number
		that prevents the division by zero. Defalut value is ``1.0e-6``.

	min_ratio : float, optional
		Minimum permitted value of normalized projections. It must be a small
		positive number that prevents negative values normalized data.
		Defalut value is ``1.0e-6``.

	max_ratio : float, optional
		Maximum permitted value of normalized projections. It must be a
		positive number. It mitigates the magnitude of the bright outliers within
		normalized data. Defalut value is ``10``.

	log : bool, optional
		If ``True`` the log-transform of the normalized data is performed. If
		``False``, the normalized data without log-transform are returned.
		Default value is ``False``.

	mode : string, optional
	 	If `dose_draw` or `crop_draw` is ``True`` the user can select interactively
		the ROI. A window showing a representative image of the projection stack
		is created. This image can be the mean or the standard deviation computed
		pixel-wise over the projection stack. Hence, allowed values of `mode` are
		``mean`` and ``std``. Default value is ``mean``.

	sino_order : bool, optional
		If ``True`` a stack of sinograms is returned (0 axis represents the
		projections y-axis). If ``False`` a stack of projections is returned
		(0 axis represents theta). Default value is ``False``.

	Returns
	-------
	out : ndarray
		Three-dimensional stack of the normalized projections.

	out_180 : 2d arrays
		The normalized projection at 180 degree. It is returned only
		if `proj_180` is an array.

	References
	----------
	.. [1] D. Micieli et al., A comparative analysis of reconstruction methods
	 applied to Neutron Tomography, Journal of Instrumentation, Volume 13,
	 June 2018.

	Examples
	--------
	Normalize dataset selecting interactively the ROI to crop and the dose ROI.

	>>> import neutompy as ntp
	>>> norm =  ntp.normalize_proj(proj, dark, flat, dose_draw=True, crop_draw=True)

	Normalize dataset and the raw projection at 180 degree:

	>>> fname  = ntp.get_image_gui('', message = 'Select raw projection at 180°...')
	>>> img180 = ntp.read_image(fname)
	>>> norm, norm_180 = ntp.normalize_proj(proj, dark, flat, proj_180=img180)

	Normalize dataset using two ImageJ .roi file to define the ROI to crop and
	the dose ROI:

	>>> norm = ntp.normalize_proj(proj, dark, flat, dose_file='./dose.roi', crop_file='./crop.roi'
							 dose_draw=False, crop_draw=False)

	Normalize the dataset with the log-transform:

	>>> norm = ntp.normalize_proj(proj, dark, flat, log=True)

	Trivial data normalization using the whole field of view and without the
	dose correction:

	>>> norm = ntp.normalize_proj(proj, dark, flat, dose_draw=False, crop_draw=False)
	"""
	if(min_ratio<=0.0):
		raise ValueError('The parameter min_ratio must be positive.')
	if(max_ratio<=0.0):
		raise ValueError('The parameter min_ratio must be positive.')

	if(max_ratio<=min_ratio):
		raise ValueError('Invalid values assigned to max_ratio and min_ratio.')

	if not (proj.ndim == 3 and dark.ndim == 3 and flat.ndim == 3):
		raise ValueError('All images stack must have three dimesions.')

	if type(proj_180) is np.ndarray:
		if (proj_180.ndim !=2):
			raise ValueError('Invalid array dimensions. The projection at 180 must be a 2D array.')
		AddProj180 = True

	if proj_180 is None:
		AddProj180 = False

	# only one DOSE roi selection check
	doseON = 0
	if (dose_file):
		doseON = doseON + 1
	if (dose_coor):
		doseON = doseON + 1
	if (dose_draw):
		doseON = doseON + 1
	if(doseON>=2):
		raise ValueError('Only one selection method of the dose ROI is allowed.')

	# only one CROP roi selection check
	cropON = 0
	if (crop_file):
		cropON = cropON + 1
	if (crop_coor):
		cropON = cropON + 1
	if (crop_draw):
		cropON = cropON + 1
	if(cropON>=2):
		raise ValueError('Only one selection method of the ROI to crop is allowed.')

	if (show_opt=='mean'):
		func_show = np.mean
	elif (show_opt=='std'):
		func_show = np.std
	else:
		raise ValueError('Not valid value of the variable show_opt. Choose "mean"\
		 or "std" to show the mean or the standard deviation computed pixel-wise over\
		 all projections.')

	# get the ROI to CROP
	if(cropON):
		# read ImageJ ROI
		if(crop_file):
			rmin_c, rmax_c, cmin_c, cmax_c = get_rect_coordinates_from_roi(crop_file)

		# read coordinate from tuple
		if (crop_coor):
			rmin_c, rmax_c, cmin_c, cmax_c = crop_coor

		# draw the roi over the image
		if (crop_draw):
			print("> Crop a ROI of the projections to reconstruct...")
			show_proj = func_show(proj, axis=0, dtype=np.float32)
			rmin_c, rmax_c, cmin_c, cmax_c = draw_ROI(show_proj, 'Select the region to use for the reconstruction...' )

		proj_c = proj[:, rmin_c:rmax_c, cmin_c:cmax_c]
		dark_c = dark[:, rmin_c:rmax_c, cmin_c:cmax_c]
		flat_c = flat[:, rmin_c:rmax_c, cmin_c:cmax_c]

		crop_roi = rmin_c, rmax_c, cmin_c, cmax_c

	else:
		proj_c = proj
		dark_c = dark
		flat_c = flat
		crop_roi = None

	# get the ROI to DOSE
	if(doseON):
		# read ImageJ ROI
		if(dose_file):
			rmin_d, rmax_d, cmin_d, cmax_d = get_rect_coordinates_from_roi(dose_file)

		# read coordinate from tuple
		if (dose_coor):
			rmin_d, rmax_d, cmin_d, cmax_d  = dose_coor

		# draw the roi over the image
		if (dose_draw):
			if not crop_draw: # to prevent double computation of std
				show_proj = func_show(proj, axis=0, dtype=np.float32)
			print("> Background ROI selection. Select a region free of the sample...")
			rmin_d, rmax_d, cmin_d, cmax_d  = draw_ROI(show_proj, 'Select a region free of the sample...' )

		ds_roi = rmin_d, rmax_d, cmin_d, cmax_d


		if(mode=='mean'):
			func = np.mean
		elif(mode=='median'):
			func = np.median
		else:
			raise ValueError('Not valid function for projecting flat and dark images along z axis.\
							Set mean or media.')

		mean_flat  = func(flat[:, rmin_d:rmax_d, cmin_d:cmax_d], axis=0).astype(np.float32)
		mean_dark  = func(dark[:, rmin_d:rmax_d, cmin_d:cmax_d], axis=0).astype(np.float32)

		min_denom   = np.float32(min_denom)
		min_ratio   = np.float32(min_ratio)
		max_ratio   = np.float32(max_ratio)

		proj_d  = proj[:, rmin_d:rmax_d, cmin_d:cmax_d]
		num_d   = ne.evaluate('proj_d - mean_dark')
		den_d   = ne.evaluate('mean_flat - mean_dark')

		# (flat - dark) dose
		D0 = np.median(den_d)

		# projection dose computed in a roi free-of-sample
		D  = np.median(num_d, axis=(1,2))

		# correction factor
		if(proj.shape[0]==1):
			k = D0/D
		else:
			k = D0/(D.reshape(proj.shape[0], 1))

	# end dose computation
	else:
		ds_roi = None

	if (proj_c.shape[0] != 1):
		print('> Normalization...')
	out = _normalize(proj_c, dark_c, flat_c, mode, min_denom,  out=out)

	if(doseON):
		if(proj.shape[0]==1):
			out = k*out
		else:
			nz, ny, nx = out.shape
			out = (k*out.reshape(nz, nx*ny)).reshape(nz, ny, nx)


	min_ratio = np.float32(min_ratio)
	max_ratio = np.float32(max_ratio)

	out = ne.evaluate('where(out>max_ratio, max_ratio, out)', out = out)
	out = ne.evaluate('where(out<min_ratio, min_ratio, out)', out = out)


	if (log):
		out = ne.evaluate('-log(out)', out=out)


	if (sino_order):
		out = out.swapaxes(0,1)

	if(AddProj180):
		proj_180 = np.expand_dims(proj_180, axis=0)

		out_180  = normalize_proj(proj_180, dark, flat,  proj_180=None, out=None,
						 dose_file='', dose_coor = ds_roi, dose_draw=False,
						 crop_file='', crop_coor = crop_roi, crop_draw=False,
						 min_denom=min_denom,  min_ratio=min_ratio, max_ratio=max_ratio,
						 mode=mode, log=log,  sino_order=sino_order)

		return out, out_180[0]

	else:
		return out



def log_transform(norm_proj, out=None):
	"""
	This function computes the minus log of an array. In trasmission CT the input\
	array must be the normalized dataset after flat-fielding correction.

	Parameters
	----------
	norm_proj : ndarray
		3D stack of projections.

	out : ndarray, optional
		Output array. If same as norm_proj, computation will be done in-place.

	Returns
	-------
	out : ndarray
		Minus-log of the input array.
	"""

	out = ne.evaluate('-log(norm_proj)', out=out)

	return out


def resample(image, transform, interpolator=sitk.sitkLinear):

    dims = image.GetSize()
    reference_image = sitk.Image(dims[0], dims[1], sitk.sitkFloat32)
    default_value = 1.0

    return sitk.Resample(image, reference_image, transform, interpolator, default_value)


def rotate_2(img, theta, interpolator=sitk.sitkLinear):

	s = sitk.GetImageFromArray(img)
	aff = sitk.AffineTransform(2)

	com = np.array(list(img.shape))*0.5

	th   = -np.deg2rad(theta)
	matr = np.array([[cos(th), sin(th)] , [-sin(th), cos(th)]])


	center = np.array(img.shape)*0.5 - 0.5
	aff.SetCenter(center[::-1])
	#shift = np.array( com - center)
	#aff.SetTranslation ( -np.dot(matr,shift))

	aff.SetMatrix(matr.flatten())

	out= resample(s, aff)
	outimg = sitk.GetArrayFromImage(out)

	return outimg


def rotate_sitk(img, theta, interpolator=sitk.sitkLinear):

	th = -np.deg2rad(theta)
	l1,l2 = img.shape
	imgpad_size = np.array([ int(round(abs(l1*cos(th)) + abs(l2*sin(th)))),
	 						int(round(abs(l2*cos(th)) + abs(l1*sin(th)))) ])


	before =  (imgpad_size - np.array(img.shape))//2
	after  =  imgpad_size -  np.array(img.shape) -before

	before = tuple(before)
	after  = tuple(after)

	imgpad = np.pad(img, pad_width=((before[0], after[0]),  (before[1], after[1])), mode='edge')


	out = rotate_2(imgpad, theta, interpolator)

	return out[ before[0]:(l1+before[0]), before[1]:(l2 + before[1])]



def find_COR(proj_0, proj_180, nroi=None, ref_proj=None, ystep=5, ShowResults=True):
	"""
	This function estimates the offset and the tilt angle of the rotation axis
	respect to the detector using the projections at 0 and at 180 degree. The user
	selects interactively different regions where the sample is visible. Once the
	position of the rotation axis is found the results are shown in two figures.
	In the first are shown the computed rotation axis and the image obtained as
	proj_0 - pro_180[:,::-1] (the projection at 180 is flipped horizontally)
	before the correction.
	The second figure shows the difference image proj_0 - pro_180[:,::-1] after
	the correction and the histogram of abs(proj_0 - pro_180[:,::-1]).

	Parameters
	----------
	proj_0 : 2d array
		The projection at 0 degrees.

	proj_180 : 2d array
		The projection at 180 degrees.

	nroi : int, optional
		The number of the region of interest to select  for the computation
		of the rotation axis position. Default is None, hence the value is read
		as input from keyboard.

	ref_proj: 2d array
		The image shown to select the region of interest. Default is None, hence
		proj_0 is shown.

	ystep: int, optional
		The center of rotation position is computed every `ystep`. Default value
		is 5.

	ShowResults: bool, optional
		If True, the the two figures that summarize the result are shown.
		Default is True.

	Returns
	-------
	middle_shift : float
		The horizontal shift of the rotation axis respect to the center of the
		detector.

	theta : float
		The tilt angle formed by the rotation axis and the vertical axis of the
		detector.
	"""
	if (ref_proj is None):
		ref_proj = proj_0

	proj_0   = proj_0.astype(np.float32)
	proj_180 = proj_180.astype(np.float32)

	# number of pixels in a row
	nd = proj_0.shape[1]
	# number of slices
	nz = proj_0.shape[0]

	# array containing that contains the z points within the rois selected by the user
	slices = np.array([], dtype=np.int32)

	print('> Finding the rotation axis position...')
	# set ROI number
	if not nroi:
		print('To compute the rotation axis position it is necessary to select one or multiple regions where the sample is present.\nHence you must draw the different regions vertically starting from top to bottom.')
		while True:
			nroi = input('> Insert the number of regions to select: ')
			if(nroi.isdigit() and int(nroi)>0):
				nroi = int(nroi)
				break
			else:
				print('Not valid input.')

	tmin = - nd//2
	tmax =   nd - nd//2


	# ROI selection
	for i in range(0, nroi):
		# print number of rois
		print('> Select ROI ' + str(i+1))
		title = 'Select region n. : ' + str(i+1)
		ymin, ymax, _, _ = draw_ROI(ref_proj, title)

		aus = np.arange(ymin, ymax +1, ystep)
		slices = np.concatenate((slices, aus), axis=0)

	shift = np.zeros(slices.size)
	proj_flip = proj_180[:, ::-1]

	# loop over different rows
	for z, slc in enumerate(slices):

		minimum = 1e7
		index_min = 0

		# loop over the shift
		for t in range(tmin, tmax + 1):

			posz = np.round(slc).astype(np.int32)
			rmse = np.square( (np.roll(proj_0[posz], t, axis = 0) - proj_flip[posz]) ).sum() / nd

			if(rmse <= minimum):
				minimum = rmse
				index_min = t

		shift[z] = index_min

	# perform linear fit
	par = np.polyfit(slices, shift, deg=1)
	m = par[0]
	q = par[1]

	# compute the tilt angle
	theta = np.arctan(0.5*m)   # in radians
	theta = np.rad2deg(theta)

	# compute the shift
	offset       = (np.round(m*nz*0.5 + q)).astype(np.int32)*0.5
	middle_shift = (np.round(m*nz*0.5 + q)).astype(np.int32)//2


	print("Rotation axis Found!")
	print("offset =", offset, "   tilt angle =", theta, "°"  )

	p0_r = np.zeros(proj_0.shape, dtype=np.float32)
	p90_r = np.zeros(proj_0.shape, dtype=np.float32)


	# plot difference proj_0 - proj_180_flipped
	plt.figure('Analysis of the rotation axis position', figsize=(14,5), dpi=96)
	plt.subplots_adjust(wspace=0.5)
	ax1 = plt.subplot(1,2,1)

	diff = proj_0 - proj_flip
	mu = np.median(diff)
	s  = diff.std()

	plt.imshow(diff, cmap='gray', vmin=mu-s, vmax=mu+s)

	info_cor = 'offset = '  + "{:.2f}".format(offset) + '\n       θ = ' + "{:.3f}".format(theta)
	anchored_text1 = AnchoredText(info_cor, loc=2)
	ax1.add_artist(anchored_text1)

	plt.title('$P_0 - P^{flipped}_{\pi}$ before correction')
	plt.colorbar(fraction=0.046, pad=0.04)


	zaxis = np.arange(0, nz)
	# plot fit
	plt.plot( nd*0.5 - 0.5*m*zaxis - 0.5*q, zaxis,'b-')
	# plot data
	plt.plot( nd*0.5 - 0.5*shift, slices, 'r.', markersize=3)
	# draw vertical axis
	plt.plot([0.5*nd, 0.5*nd], [0, nz-1], 'k--' )

	# show results of the fit
	ax2 = plt.subplot(1,2,2)
	info_fit = 'shift = '  + "{:.3f}".format(m) + '*y + ' + "{:.3f}".format(q)
	anchored_text2 = AnchoredText(info_fit, loc=9)
	plt.plot(zaxis, m*zaxis + q, 'b-', label = 'fit')
	plt.plot(slices, shift, 'r.', label='data')
	plt.xlabel('$y$')
	plt.ylabel('shift')
	plt.title('Fit result')
	ax2.add_artist(anchored_text2)

	plt.legend()


	#~ p0_r = np.roll(rotate(proj_0, theta, preserve_range=True,order=0, mode='edge'),    middle_shift , axis=1)
	#~ p90_r = np.roll(rotate(proj_180, theta, preserve_range=True, order=0, mode='edge'),  middle_shift, axis=1)
	p0_r = np.roll(rotate_sitk(proj_0, theta, interpolator=sitk.sitkLinear),    middle_shift , axis=1)
	p90_r = np.roll(rotate_sitk(proj_180, theta, interpolator=sitk.sitkLinear),  middle_shift, axis=1)


	# FIGURE with difference image and histogram
	plt.figure('Results of the rotation axis correction', figsize=(14,5), dpi=96)

	plt.subplot(1,2,1)
	plt.subplots_adjust(wspace=0.5)

	diff2 = p0_r - p90_r[:,::-1]
	mu = np.median(diff2)
	s  = diff2.std()
	plt.imshow(diff2 , cmap='gray', vmin=mu-s, vmax=mu+s)
	plt.title('$P_0 - P^{flipped}_{\pi}$ after correction')
	plt.colorbar(fraction=0.046, pad=0.04)



	# histogram of squares of residuals
	ax3 = plt.subplot(1,2,2)
	nbins = 1000
	row_marg = int(0.1*diff2.shape[0])
	col_marg = int(0.1*diff2.shape[1])
	absdif = np.abs(diff2)[row_marg:-row_marg, col_marg:-col_marg]
	[binning, width] = np.linspace(absdif.min(), absdif.max(), nbins, retstep=True)
	cc, edge = np.histogram(absdif, bins=binning)
	plt.bar(edge[:-1]+width*0.5, cc, width, color='C3', edgecolor='k', log=True)
	plt.gca().set_xscale("log")
	plt.gca().set_yscale("log")
	plt.xlim([0.01, absdif.max()])
	plt.xlabel('Residuals')
	plt.ylabel('Entries')
	plt.title('Histogram of residuals')

	# write text about residuals
	res =  np.abs(diff2).mean()
	info_res = '$||P_0 - P^{flipped}_{\pi}||_1 / N_{pixel}$ = ' + "{:.4f}".format(res)
	anchored_text3 = AnchoredText(info_res, loc=1)
	ax3.add_artist(anchored_text3)

	print("average of residuals  = ", res)


	if(ShowResults):
		plt.show(block=False)

	return middle_shift, theta


def correction_COR(norm_proj, proj_0, proj_180, show_opt='mean', shift=None,
				theta=None, nroi=None, ystep=5):
	"""
	This function corrects the misalignment of the rotation axis respect to the
	vertical axis of the detector. The user can choose to insert manually the offset
	and the tilt angle of the rotation axis	respect to the detector, if known
	parameters, or to estimate them using the projections at 0 and at 180 degrees.
	In this case, the user selects interactively different regions where the
	sample is visible. Once the position of the rotation axis is found the
	results are shown in two figures. In the first are shown the estimated rotation
	axis and the image obtained as proj_0 - pro_180[:,::-1] (projection at 180
	is flipped horizontally) before the correction.
	The second figure shows the difference image (proj_0 - pro_180[:,::-1]) after
	the correction and the histogram of abs(proj_0 - pro_180[:,::-1]).

	Parameters
	----------
	norm_proj : ndarray
		A stack of projections. The 0-axis represents theta.

	proj_0 : 2d array
		The projection at 0 degrees.

	proj_180 : 2d array
		The projection at 180 degrees.

	show_opt : str, optional
		A string defining the image to show for the selection of the ROIs.

		Allowed values are:

		``mean`` -> shows the mean of `norm_proj` along the O-axis.

		``std``  -> shows the standard deviation of `norm_proj` along the O-axis

		``zero`` -> shows proj_0

		``pi``   -> shows proj_180

		Default value is ``mean``.

	shift: int, optinal
		The horizontal shift in pixel of the rotation axis respect to the
		vertical axis of the detector. It can be specified if known parameter.
		The default is None, hence this parameter is estimated from the projections.

	theta: int, optinal
		The tilt angle in degrees of the rotation axis respect to the vertical
		axis of the detector. It can be specified if known parameter.
		The default is None, hence this parameter is estimated from the projections.

	nroi : int, optional
		The number of the region of interest to select  for the computation
		of the rotation axis position. Default is None, hence the value is read
		as input from keyboard.

	ystep: int, optional
		The center of rotation position is computed every `ystep`. Default value
		is 5.

	Returns
	-------
	norm_proj : ndarray
		The stack after the correction. The computation is done in place.
	"""

	# compute COR axis interactively, if shift and theta are not known
	if (shift==None and theta==None):

		if (show_opt == 'mean'):
			proj2show = norm_proj.mean(axis=0, dtype=np.float32)
		elif (show_opt == 'std'):
			proj2show = norm_proj.std(axis=0, dtype=np.float32)
		elif (show_opt == 'zero'):
			proj2show = proj_0
		elif (show_opt == 'pi'):
			proj2show = proj_180
		else:
			raise ValueError('Flag show_opt not valid. Allowed values are `mean`, `std`, `zero` or `pi`.')

		condition = True

		while condition:

			shift, theta = find_COR(proj_0, proj_180, nroi=nroi, ref_proj=proj2show, ystep=ystep, ShowResults=True)
			sleep(0.5)

			while True:
					ans = input('> Rotation axis found. Do you want to correct all projections?\
					 \n[Y] Yes, correct them.  \n[N] No, find it again.\
					 \n[C] Cancel and abort the script.\
					 \nType your answer and press [Enter] :')
					if(ans=='Y' or ans=='y'):
						plt.close('all')
						condition = False
						break
					elif(ans=='N' or ans=='n'):
						plt.close('all')
						break
					elif(ans=='C' or ans=='c'):
						plt.close('all')
						print('> Script aborted.')
						sys.exit()
					else:
						print('Input not valid.')

	# end if

	# correct all projections
	print('> Correcting rotation axis misalignment...')
	for s in tqdm(range(0, norm_proj.shape[0]), unit=' images'):
		# norm_proj[s,:,:] = np.roll(rotate(norm_proj[s,:,:], theta, preserve_range=True, order=1, mode='edge'), shift, axis=1)
		norm_proj[s,:,:] = np.roll(rotate_sitk(norm_proj[s,:,:], theta, interpolator=sitk.sitkLinear), shift, axis=1)

	return norm_proj


def median_filter(img, radius):
	"""
	This function applies a median filter on an image.

	Parameters
	----------
	img : 2d array
		The image to filter. It must be a two-dimensional array.

	radius: int or tuple of int
		The radius of the neighborhood. The radius is defined separately for
		each dimension as the number of pixels that the neighborhood extends
		outward from the center pixel.

	Returns
	-------
		The filtered image.
	"""
	if (img.ndim != 2):
		raise ValueError('The input iamge must be two-dimensional.')

	simg = sitk.GetImageFromArray(img)
	mf = sitk.MedianImageFilter()
	mf.SetRadius(radius)
	fimg = sitk.GetArrayFromImage(mf.Execute(simg))

	return fimg

def median_filter_stack(arr, radius, axis=0):
	"""
	This function applies a 2D median filter on a stack of images.

	Parameters
	----------
	img : ndarray
		The stack to filter. It must be a three-dimensional array.

	radius: int or tuple of int
		The radius of the neighborhood. The radius is defined separately for
		each dimension as the number of pixels that the neighborhood extends
		outward from the center pixel.

	axis : int
		The axis along which the 2D median filter is iterated.

	Returns
	-------
	out : ndarray
		The filtered stack of images.
	"""
	if (arr.ndim != 3):
		raise ValueError('The input array must be three-dimensional.')

	arr = arr.swapaxes(0, axis)

	for i in tqdm(range(0, arr.shape[0]), unit=' images'):
		arr[i] = median_filter(arr[i], radius)

	arr = arr.swapaxes(0, axis)

	return arr


def convolution_2D (img, kernel, boundary='ZERO_PAD', out_mode='SAME'):
	"""
	Perform a 2D convolution with an arbitrary kernel.

	Parameters
	----------
	img : 2d array
		The input image.

	kernel : 2d array
		The kernel to use for the convolution.

	boundary: str
		The boundary padding type. Allowed values are:
		``ZERO_PAD``, ``ZERO_FLUX_NEUMANN_PAD`` and ``PERIODIC_PAD``.

	out_mode: str
		Sets the output region mode. If set to `SAME`, the output region
		will be the same as the input region, and regions of the image
		near the boundaries will contain contributions from outside the
		input image as determined by the boundary condition set in
		`boundary`. If set to `VALID`, the output region
		consists of pixels computed only from pixels in the input image
		(no extrapolated contributions from the boundary condition are
		needed). The output is therefore smaller than the input
		region. Default value is `SAME`.

	Returns
	-------
	out : 2d array
		The filtered image. The type of the array is inferred
		from img and kernel type.
	"""

	simg    = sitk.GetImageFromArray(img)
	skernel = sitk.GetImageFromArray(kernel)

	conv = sitk.ConvolutionImageFilter()
	conv.SetBoundaryCondition(eval('conv.' + boundary))
	conv.SetOutputRegionMode(eval('conv.' + out_mode))

	filt = sitk.GetArrayFromImage(conv.Execute(simg, skernel))

	return filt

def std_map(img, radius):

	if (type(radius) is not int or radius==0):
		raise ValueError('Kernel radius must be a positive integer.')

	vartype = img.dtype # get the array type
	img     = img.astype(vartype)
	img2    = img**2
	ones    = np.ones(img.shape, dtype=vartype)
	kernel = np.ones((2*radius+1, 2*radius+1),  dtype=vartype)

	s  = convolution_2D(img, kernel)
	s2 = convolution_2D(img2, kernel)
	ns = convolution_2D(ones, kernel)

	return np.sqrt((s2 - s**2 / ns) / (ns-1))


def remove_outliers(img, radius, threshold, outliers='bright', k=1.0, out=None):
	"""
	This function removes bright and dark outliers from an image.
	It replaces a pixel by the median of the pixels in the neighborhood
	if it deviates from the median by more than a certain value (k*threshold).
	The threshold can be specified by the user as a global value or computed
	proportionally to the local standard deviation of the projection.

	Parameters
	----------
	img : 2d array
		The image to elaborate.

	radius: int or tuple of int
		The radius of the neighborhood. The radius is defined separately for
		each dimension as the number of pixels that the neighborhood extends
		outward from the center pixel.

	threshold: float or str
		If it is the string 'local' the local standard deviation map is taken
		into account. Conversely, if it is a float number the threshold is global.

	outliers: str, optional
		A string defining the type of outliers to remove. Allowed values are
		``bright`` and ``dark``. Default is `bright`.

	k : float, optional
		A pixel is replaced by the median of the pixels in the neighborhood
		if it deviates from the median by more than k*threshold.
		Default value is 1.0.

	out: 2d array, optional
		If same as img, then the computation is done in place. Default is None,
		hence this behaviour is disabled.

	Returns
	-------
	out : 2d array
		The image obtained by removing the outliers.
	"""
	if not (threshold=='local' or type(threshold) in [float, int]):
		raise ValueError('The threshold must be a float variable or the string `local`.')

	if(threshold=='local'):
		threshold = std_map(img, radius)

	median = median_filter(img, radius)

	if(outliers=='bright'):
		diff    = img - median
		spot    = diff > threshold * k
	elif(outliers=='dark'):
		diff    =  median - img
		spot    = diff   > threshold * k

	# convert type
	cut = np.array(threshold*k, dtype = img.dtype)

	out = ne.evaluate('where(diff>cut, median, img)', out=out)

	return out


def remove_outliers_stack(arr, radius, threshold, axis=0, outliers='bright', k=1.0, out=None):
	"""
	This function removes bright and dark outliers from a stack of images.
	The algorithm elaborates 2d images and the filtering is iterated over all
	images in the stack.
	The function replaces a pixel by the median of the pixels in the 2d neighborhood
	if it deviates from the median by more than a certain value (k*threshold).
	The threshold can be specified by the user as a global value or computed
	proportionally to the local standard deviation of the projection.

	Parameters
	----------
	arr : ndarray
		The stack to elaborate.

	radius: int or tuple of int
		The radius of the 2D neighborhood. The radius is defined separately for
		each dimension as the number of pixels that the neighborhood extends
		outward from the center pixel.

	threshold: float or str
		If it is the string 'local' the local standard deviation map is taken
		into account. Conversely, if it is a float number the threshold is global.

	axis : int
		The axis along wich the outlier removal is iterated.

	outliers: str, optional
		A string defining the type of outliers to remove. Allowed values are
		``bright`` and ``dark``. Default is `bright`.

	k : float, optional
		A pixel is replaced by the median of the pixels in the neighborhood
		if it deviates from the median by more than k*threshold.
		Default value is 1.0.

	out: 2d array, optional
		If same as arr, then the computation is done in place. Default is None,
		hence this behaviour is disabled.

	Returns
	-------
	out : ndarray
		The array obtained by removing the outliers.
	"""
	arr = arr.swapaxes(0, axis)

	if(out is None):
		out = arr

	print('> Removing ' + outliers + ' outliers...')
	for i in tqdm(range(0, arr.shape[0]), unit=' images'):
		remove_outliers(arr[i], radius, threshold, outliers, k, out=out[i])

	out = out.swapaxes(0, axis)
	return out


def remove_stripe(img, level, wname='db5', sigma=1.5):
	"""
	Suppress horizontal stripe in a sinogram using the Fourier-Wavelet based
	method by Munch et al. [2]_.

	Parameters
	----------
	img : 2d array
		The two-dimensional array representig the image or the sinogram to de-stripe.

	level : int
		The highest decomposition level.

	wname : str, optional
		The wavelet type. Default value is ``db5``

	sigma : float, optional
		The damping factor in the Fourier space. Default value is ``1.5``

	Returns
	-------
	out : 2d array
		The resulting filtered image.

	References
	----------
	.. [2] B. Munch, P. Trtik, F. Marone, M. Stampanoni, Stripe and ring artifact removal with
		   combined wavelet-Fourier filtering, Optics Express 17(10):8567-8591, 2009.
	"""

	nrow, ncol = img.shape

	# wavelet decomposition.
	cH = []; cV = []; cD = []

	for i in range(0, level):
		img, (cHi, cVi, cDi) = pywt.dwt2(img, wname)
		cH.append(cHi)
		cV.append(cVi)
		cD.append(cDi)

	# FFT transform of horizontal frequency bands
	for i in range(0, level):
		# FFT
		fcV=fftshift(fft(cV[i], axis=0))
		my, mx = fcV.shape

		# damping of vertical stripe information
		yy2  = (np.arange(-np.floor(my/2), -np.floor(my/2) + my))**2
		damp = - np.expm1( - yy2 / (2.0*(sigma**2)) )
		fcV  = fcV * np.tile(damp.reshape(damp.size, 1), (1,mx))

		#inverse FFT
		cV[i] = np.real( ifft( ifftshift(fcV), axis=0) )

	# wavelet reconstruction
	for i in  range(level-1, -1, -1):
		img = img[0:cH[i].shape[0], 0:cH[i].shape[1]]
		img = pywt.idwt2((img, (cH[i], cV[i], cD[i])), wname)

	return img[0:nrow, 0:ncol]


def remove_stripe_stack(arr, level, wname='db5', sigma=1.5, axis=1, out=None):
	"""
	Suppress horizontal stripe in a stack of sinograms or a stack of projections
	using the Fourier-Wavelet based method by Munch et al. [3]_ .

	Parameters
	----------
	arr : 3d array
		The tomographic data. It can be a stack of projections (theta is the 0-axis)
		or a stack of images (theta is the 1-axis).

	level : int
		The highest decomposition level

	wname : str, optional
		The wavelet type. Default value is ``db5``

	sigma : float, optional
		The damping factor in the Fourier space. Default value is ``1.5``

	axis : int, optional
		The axis index of the theta axis.
		Default value is ``1``.

	out : None or ndarray, optional
		The output array returned by the function. If it is the same as `arr`,
		the computation will be done in place.
		Default value is None, hence a new array is allocated and returned by
		the function.

	Returns
	-------
	outarr : 3d array
		The resulting filtered dataset.

	References
	----------
	.. [3] B. Munch, P. Trtik, F. Marone, M. Stampanoni, Stripe and ring artifact removal with
		   combined wavelet-Fourier filtering, Optics Express 17(10):8567-8591, 2009.
	"""

	if(arr.ndim != 3):
		raise ValueError('Input array must be three-dimensional')

	if(out is None):
		out_arr = np.zeros(arr.shape, dtype=arr.dtype)

	if isinstance(out, np.ndarray):
		out_arr = out

	arr     = arr.swapaxes(0, axis)
	out_arr = out_arr.swapaxes(0,axis)

	print('Removing stripe...')
	for i in tqdm(range(0, arr.shape[0]), unit='images'):
		out_arr[i] = remove_stripe(arr[i], level, wname, sigma)


	out_arr = out_arr.swapaxes(0,axis)

	return out_arr



#-----------------------------------------------------------------------
#Copyright 2013 Centrum Wiskunde & Informatica, Amsterdam
#
#Author: Daniel M. Pelt
#Contact: D.M.Pelt@cwi.nl
#Website: http://dmpelt.github.io/pynnfbp/
#
#
#This file is part of the PyNN-FBP, a Python implementation of the
#NN-FBP tomographic reconstruction method.
#
#PyNN-FBP is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.
#
#PyNN-FBP is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with PyNN-FBP. If not, see <http://www.gnu.org/licenses/>.
#
#-----------------------------------------------------------------------

import tempfile
import numpy as np

import scipy.io as sio

import sys
import os
import math
import random
class TrainingData(object):
    '''Base object of a class that represents training or validation data used
    during training of a network.
    
    An implementing class should define ``getDataBlock``, ``addDataBlock`` and ``normalizeData`` methods. See,
    for example, :class:`HDF5TrainingData`.
    
    :param data: Dataset to pick pixels from. (see :mod:`nnfbp.DataSet`)
    :type data: DataSet
    :param nPoints: Number of pixels to pick.
    :type nPoints: :class:`int`
    :param blockSize: Size of each data block.
    :type blockSize: :class:`int`
    '''
    
    def __setupIDX(self,size):
        '''Create a variable ``idx`` that gives location of pixels that can be picked.'''
        ym,xm = np.ogrid[-(size-1.)/2.:(size-1.)/2.:complex(0,size),-(size-1.)/2.:(size-1.)/2.:complex(0,size)]
        bnd = (size)**2/4
        self.mask = xm**2+ym**2<=bnd
        x,y = np.where(self.mask==True)
        self.idx = zip(x,y)
    
    def __getPickedIndices(self,nToPick):
        '''Return a list of the location of ``nToPick`` randomly selected pixels.'''
        nTimesToDo = int(math.ceil(nToPick/float(len(self.idx))))
        iList = []
        for i in range(nTimesToDo):
            iList.extend(self.idx)
        return zip(*random.sample(iList,nToPick))
    
    def __getPickedIndicesWithMask(self,nToPick,mask):
        '''Return a list of the location of ``nToPick`` randomly selected pixels.'''
        maskCombined = self.mask+mask
        x,y = np.where(maskCombined>1)
        idx = zip(x,y)
        nTimesToDo = int(math.ceil(nToPick/float(len(idx))))
        iList = []
        for i in range(nTimesToDo):
            iList.extend(idx)
        return zip(*random.sample(iList,nToPick))
    
    def __init__(self,data,nPoints,network,blockSize=10000):
        tmpFl = tempfile.mkstemp(dir=network.tmpDir)
        self.fn= tmpFl[1]
        os.close(tmpFl[0])
        pickArray = np.histogram(np.floor(data.nImages*np.random.rand(nPoints)), data.nImages, (0,data.nImages))[0]
        if np.max(pickArray)>blockSize:
            raise Exception('Buffer size is too small!')
        nParameters = network.red.outSize
        self.nPar = nParameters
        curData = np.empty((np.max(pickArray),nParameters+1))
        outData = np.empty((blockSize,nParameters+1))
        self.__setupIDX(network.proj.recSize)
        self.nBlocks=0
        nInBlock=0
        i=0
        for i in range(len(data)):
            example = data[i]
            nToPick = pickArray[i]
            
            if nToPick==0: continue
            image = example[0]
            sino = example[1]
            angles = example[2]
            if len(example)>3:
                pickedIndices = self.__getPickedIndicesWithMask(nToPick,example[3])
            else:
                pickedIndices = self.__getPickedIndices(nToPick)
            for j in range(nParameters):
                backImage = network.proj.reconstructWithFilter(sino,network.red.filters[:,j])
                curData[:nToPick,j] = backImage[pickedIndices]
            curData[:nToPick,nParameters] = image[pickedIndices]
            if nInBlock+nToPick<blockSize:
                outData[nInBlock:nInBlock+nToPick,:]=curData[:nToPick,:].copy()
                nInBlock+=nToPick
            else:
                nToWrite = blockSize-nInBlock
                nLeft = nToPick - nToWrite
                if nToWrite>0:
                    outData[nInBlock:blockSize,:] = curData[0:nToWrite,:].copy()
                    self.addDataBlock(outData,self.nBlocks)
                    self.nBlocks+=1
                nInBlock=0
                if nLeft>0:
                    outData[0:nLeft,:] = curData[nToWrite:nToPick,:].copy()
                    nInBlock+=nLeft
            percDone = float(blockSize*self.nBlocks + nInBlock)/nPoints
            nTicksDone = (int)(percDone*60)
            sys.stdout.write('\r[%s>%s] %d%% %s' % ('-'*nTicksDone, ' '*(60-nTicksDone), 100*percDone,50*' '))             
            sys.stdout.flush()
        if nInBlock>0:
            self.addDataBlock(outData[0:nInBlock,:], self.nBlocks)
            self.nBlocks+=1
        sys.stdout.write('\n')
        sys.stdout.flush()
    
    def addDataBlock(self,data,i):
        '''Add a block of data to the set.
        
        :param data: Block of data to add.
        :type data: :class:`numpy.ndarray`
        :param i: Position to add block to.
        :type i: :class:`int`
        '''
        raise NotImplementedError("TrainingData: Subclass should implement this method.")
    
    def getDataBlock(self,i):
        '''Get a block of data from the set.
        
        :param i: Position of block to get.
        :type i: :class:`int`
        :returns: :class:`numpy.ndarray` -- Block of data.
        '''
        raise NotImplementedError("TrainingData: Subclass should implement this method.")
    
    def getMinMax(self):
        '''Returns the minimum and maximum values of each column of the entire set.
        
        :returns: :class:`tuple` with:
        - ``minL`` -- :class:`numpy.ndarray` with minimum value of each column except last.
        - ``maxL`` -- :class:`numpy.ndarray` with maximum value of each column except last.
        - ``minIn`` -- :class:`float` minimum values of last column.
        - ``maxIn`` -- :class:`float` maximum values of last column.
        '''
        minL = np.empty(self.nPar)
        minL.fill(np.inf)
        maxL = np.empty(self.nPar)
        maxL.fill(-np.inf)
        maxIn = -np.inf
        minIn = np.inf
        for i in range(self.nBlocks):
            data = self.getDataBlock(i)
            if data is None:
                continue
            maxL = np.maximum(maxL, data[:, 0:self.nPar].max(0))
            minL = np.minimum(maxL, data[:, 0:self.nPar].min(0))
            maxIn = np.max([maxIn, data[:, self.nPar].max()])
            minIn = np.min([minIn, data[:, self.nPar].min()])
        return (minL,maxL,minIn,maxIn)
    
    def normalizeData(self,minL,maxL,minIn,maxIn):
        '''Normalize the set such that every column is in range (0,1), except for the last column,
        which will be normalized to (0.25,0.75). Parameters are like ``getMinMax()``.
        '''
        raise NotImplementedError("TrainingData: Subclass should implement this method.")
    
    def close(self):
        '''Close the underlying file.'''
        os.remove(self.fn)

class MATTrainingData(TrainingData):
    '''Implementation of :class:`TrainingData` that uses a MAT files to store data.
    '''
    
    def getDataBlock(self,i):
        if not self.tileM is None:
            data = sio.loadmat(self.fls[i])[self.dataname]
            data[:,0:self.nPar] =2*(data[:,0:self.nPar]-self.tileM)/self.maxmin - 1
            data[:,self.nPar] = 0.25+(data[:,self.nPar]-self.minIn)/(2*(self.maxIn-self.minIn))
            return data
        else:
            return sio.loadmat(self.fls[i])[self.dataname]
    
    def normalizeData(self,minL,maxL,minIn,maxIn):
        data = sio.loadmat(self.fls[0])[self.dataname]
        self.tileM = np.tile(minL, (data.shape[0],1))
        self.maxmin = np.tile(maxL-minL, (data.shape[0],1))
        self.minIn = minIn
        self.maxIn = maxIn
    
    def __init__(self,fls,dataname='mat'):
        self.fls = fls
        self.nBlocks = len(fls)
        self.dataname = dataname
        self.tileM = None
        self.nPar = (sio.loadmat(self.fls[0])[self.dataname]).shape[1]-1



#-----------------------------------------------------------------------
#Copyright 2015 Daniel M. Pelt
#
#Contact: D.M.Pelt@cwi.nl
#Website: http://www.dmpelt.com
#
#
#This file is part of the PyMR-FBP, a Python implementation of the
#MR-FBP tomographic reconstruction method.
#
#PyMR-FBP is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.
#
#PyMR-FBP is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with PyMR-FBP. If not, see <http://www.gnu.org/licenses/>.
#
#-----------------------------------------------------------------------
import astra
import numpy as np
from six.moves import range
from scipy.signal import fftconvolve
import scipy.ndimage.filters as snf
import scipy.io as sio
import numpy.linalg as na

import os, errno

try:
    from pywt import wavedec2
except:
    astra.log.warn("No pywavelets installed, wavelet functions will not work.")

def mkdir_p(path):
    try:
        os.makedirs(path)
    except OSError as exc: # Python >2.5
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else: raise


class plugin(astra.plugin.ReconstructionAlgorithm2D):
    """Reconstructs using the MR-FBP method [1].

    Options:

    'save_filter' (optional): file to save filter to (overwrites existing)
    'use_saved_filter' (optional): file to load filter from
    'reg_grad' (optional): amount of l2 gradient minimization
    'reg_path' (optional): folder to save range of 'reg_grad' values to
    'reg_range' (optional): range of 'reg_grad' values to try (of form (min, max, number of trials))
    'reg_wav' (optional): amount of l2 wavelet minimization
    'wav_bas' (optional): wavelet to use (see pywt.wavelist())
    'nlinear' (optional): number of linear steps in exponential binning
    
    [1] Pelt, D. M., & Batenburg, K. J. (2014). Improving filtered backprojection 
        reconstruction by data-dependent filtering. Image Processing, IEEE 
        Transactions on, 23(11), 4750-4762.
    """
    
    astra_name="MR-FBP"
    
    def customFBP(self, f, s):
        sf = np.zeros_like(s)
        padded = np.zeros(s.shape[1]*2)
        l = int(s.shape[1]/2.)
        r = l+s.shape[1]
        bl = f.shape[0]/len(s)
        for i in range(sf.shape[0]):
            padded[l:r] = s[i]
            padded[:l] = padded[l]
            padded[r:] = padded[r-1]
            sf[i] = fftconvolve(padded,f,'same')[l:r]
        return (self.W.T*sf).reshape(self.v.shape)
    
    def initialize(self, cfg, nlinear=2, reg_wav=None, wav_bas='haar', reg_grad=None, save_filter=None, use_saved_filter=None, reg_path=None, reg_range=(1,100,10)):
        self.W = astra.OpTomo(cfg['ProjectorId'])
        self.reg_gr = reg_grad
        self.reg_wav = reg_wav
        self.save_filter=save_filter
        self.use_saved=use_saved_filter
        self.reg_path = reg_path
        self.reg_range = reg_range
        self.wav_bas = wav_bas
        
        if not self.use_saved:
            self.bck = (self.W.T*np.ones_like(self.s)<self.s.shape[0]-0.5).reshape(self.v.shape)
        
        if self.reg_path:
            self.reg_gr=1.
        
        fs = self.s.shape[1]
        if fs%2==0:
            fs += 1
        mf = int(fs/2)
        
        w=1
        c=mf
        
        bas = np.zeros(fs,dtype=np.float32)
        self.basis = []
        count=0
        while c<fs:
            bas[:]=0
            l = c
            r = c+w
            if r>fs: r=fs
            bas[l:r]=1
            if l!=0:
                l = fs-c-w
                r = l+w
                if l<0: l=0
                bas[l:r]=1
            self.basis.append(bas.copy())
            c += w
            count += 1
            if count>nlinear:
                w=2*w
        self.nf = len(self.basis)

    def run(self, iterations):
        if self.use_saved:
            flt = sio.loadmat(self.use_saved)['mrfbp_flt'].flatten()
            self.v[:] = self.customFBP(flt, self.s)
            return
        nrows = self.s.size
        ncols = self.nf
        if self.reg_gr:
            nrows += 2*self.v.size
        if self.reg_wav:
            q = wavedec2(self.v,self.wav_bas)
            l = [q[0].flatten()]
            for z in range(1,len(q)):
                l.extend([y.flatten() for y in q[z]])
            l = np.hstack(l)
            nrows += l.size
        A = np.zeros((nrows,ncols),dtype=np.float32)
        astra.log.info("Generating MR-FBP matrix")
        for i, bas in enumerate(self.basis):
            astra.log.info('{:.2f} % done'.format(100*float(i)/self.nf))
            img = self.customFBP(bas, self.s)
            if self.reg_wav:
                q = wavedec2(img,self.wav_bas)
                l = [q[0].flatten()]
                for z in range(1,len(q)):
                    l.extend([y.flatten() for y in q[z]])
                l = np.hstack(l)
                A[nrows-l.size:,i] = self.reg_wav*l
            if self.reg_gr!=None:
                dx = np.zeros_like(self.v)
                dx[:-1,:] = img[:-1,:] - img[1:,:]
                dy = np.zeros_like(self.v)
                dx[:,:-1] = img[:,:-1] - img[:,1:]
                dx[self.bck]=0
                dy[self.bck]=0
                img[self.bck]=0
                A[0:self.s.size,i] = self.W*img
                A[self.s.size+0*self.v.size:self.s.size+1*self.v.size,i] = self.reg_gr*dx.flatten()
                A[self.s.size+1*self.v.size:self.s.size+2*self.v.size,i] = self.reg_gr*dy.flatten()
            else:
                img[self.bck]=0
                astra.extrautils.clipCircle(img)
                A[:self.s.size,i] = self.W*img
        b = np.zeros(nrows,dtype=np.float32)
        b[:self.s.size] = self.s.flatten()
        if self.reg_path:
            import tifffile
            mkdir_p(self.reg_path)
            l, r, s = self.reg_range
            astra.log.info("Generating regularized images")
            for i, reg in enumerate(np.linspace(l,r,s)):
                astra.log.info('{:.2f} % done'.format(100*float(i)/s))
                A[self.s.size:nrows-self.v.size,:] *= reg
                out = na.lstsq(A,b, rcond=-1)
                flt = np.zeros_like(self.basis[0])
                for i, bas in enumerate(self.basis):
                    flt += out[0][i]*bas
                rc = self.customFBP(flt, self.s)
                tifffile.imsave(self.reg_path + os.sep + str(reg) + '.tiff', rc)
                A[self.s.size:nrows-self.v.size,:] /= reg
            return
        else:
            out = na.lstsq(A,b, rcond=-1)
        bl = self.basis[0].shape[0]
        flt = np.zeros_like(self.basis[0])
        for i, bas in enumerate(self.basis):
            flt += out[0][i]*bas
        if self.save_filter:
            sio.savemat(self.save_filter, {'mrfbp_flt':flt}, do_compression=True, appendmat=False)
        astra.log.info("Done!")
        rc = self.customFBP(flt, self.s)
        self.v[:] = rc



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()


exec(open('neutompy/version.py').read())

setuptools.setup(
    name="neutompy",
    version=__version__,
    author="Davide Micieli",
    author_email="neutompy@gmail.com",
    description="Python package for tomographic data processing and reconstruction",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/dmici/NeuTomPy-toolbox",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Education",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Operating System :: OS Independent",
    ],
)



#-----------------------------------------------------------------------
#Copyright 2014 Daniel M. Pelt
#
#Contact: D.M.Pelt@cwi.nl
#Website: http://www.dmpelt.com
#
#
#This file is part of the PyMR-FBP, a Python implementation of the
#MR-FBP tomographic reconstruction method.
#
#PyMR-FBP is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.
#
#PyMR-FBP is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with PyMR-FBP. If not, see <http://www.gnu.org/licenses/>.
#
#-----------------------------------------------------------------------

from .astra_plugin import plugin


import numpy as np
import numexpr as ne
import logging

__author__  = "Davide Micieli"
__all__     = ['convert_to_uint']


logs = logging.getLogger(__name__)

def convert_to_uint(arr, nbit=16, down=None, up=None, crop_roi=None):
	"""
	This function converts a floating point array to 8 or 16 bit unsigned integer array.
	The resampling can be performed in a user-specified dynamic range.
	The array can be also cropped by specifying the ROI coordinates.

	Parameters
	----------
	arr : ndarray
		The floating point array to convert to unsigned integer array.

	nbit : int
		The number of bits of the returned array. In must be `8` or `16`, for
		8-bit or 16-bit unsigned integer, respectively. Default value is `16`.

	down: float, optional
		The lower range limit. If `None`, the lower limit is computed as the
		minimum value of the input array.

	up: float, optional
		The upper range limit. If `None`, the upper limit is computed as the
		maximum value of the input array.

	crop_roi: tuple, optional
		Tuple defining the ROI of the array to crop. E.g. for a 3D stack
		(smin, smax, rmin, rmax, cmin, cmax), while for a 2D image
		(rmin, rmax, cmin, cmax). Default value is `None`, which disables
		the crop of the ROI.

	Returns
	-------
	m : ndarray
		The array resampled to 8-bit or 16-bit unsigned integer.
	"""

	if(arr.ndim != 3 and  arr.ndim != 2):
		raise ValueError('The input array must have 2 or 3 dimensions.')

	# check number of bits
	if(nbit == 8):
		tp  = np.uint8
	elif(nbit == 16):
		tp  = np.uint16
	else:
		raise ValueError('The output array type must be 8 bit or 16 bit unsigned integer.')


	if(crop_roi):

		old_shape = arr.shape

		if(arr.ndim == 3):
			smin, smax, rmin, rmax, cmin, cmax = crop_roi
			arr = arr[smin:smax, rmin:rmax, cmin:cmax].astype(np.float32)

		if(arr.ndim == 2):
			rmin, rmax, cmin, cmax = crop_roi
			arr = arr[rmin:rmax, cmin:cmax].astype(np.float32)

		logs.debug('Array cropped: original shape = %s, cropped array shape = %s', old_shape, arr.shape)

	arr = arr.astype(np.float32)

	if(down == None):
		down = arr.min()

	if(up == None):
		up = arr.max()

	down = down.astype(np.float32)
	up   = up.astype(np.float32)

	factor = np.float32(2.0**nbit - 1.0)

	arr = np.clip(arr, a_min=down, a_max=up, out=arr)
	arr = ne.evaluate('factor*(arr - down)/(up - down)', out=arr, truediv=True)

	logs.info('Array converted to %s using the dynamic range %s - %s', tp, down, up)

	return (np.round(arr)).astype(dtype=tp)
